"""This module provides utilities methods."""

import copy
import os
from pathlib import Path
from causalflow.basics.constants import *


def cls():
    """Clear terminal."""
    os.system('cls' if os.name == 'nt' else 'clear')


def get_selectorpath(resfolder):
    """
    Return log file path.

    Args:
        resfolder (str): result folder.

    Returns:
        (str): log file path.
    """
    Path(resfolder).mkdir(parents=True, exist_ok=True)
    return SEP.join([resfolder, LOG_FILENAME]), SEP.join([resfolder, RES_FILENAME]), SEP.join([resfolder, DAG_FILENAME]), SEP.join([resfolder, TSDAG_FILENAME])


def create_results_folder(resfolder):
    """
    Create results folder if doesn't exist.

    Args:
        resfolder (str): result folder's name.
    """
    Path(resfolder).mkdir(parents=True, exist_ok=True)
    
    
def remove_from_list(list: list(), item) -> list():
    """
    Create a copy of a list and remove from it an item.

    Args:
        list (list): list.
        item (any): item to remove.

    Returns:
        list: new list without the item.
    """
    tmp = copy.deepcopy(list)
    tmp.remove(item)
    return tmp



#!/usr/bin/env python
"""
Implements the interactive variants of the ArcDiagram class, including the

  - InteractiveArcDiagram,
  - MutableArcDiagram, and
  - EditableArcDiagram classes,

as well as the following helper classes:

  - DraggableArcDiagram, and
  - NascentEdge.

"""

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.patches import ConnectionStyle

from ._arcdiagram_classes import ArcDiagram
from ._interactive_graph_classes import (
    DraggableGraph,
    EmphasizeOnHoverGraph,
    AnnotateOnClickGraph,
    TableOnClickGraph,
    MutableGraph,
    EditableGraph,
)
from ._utils import (
    _bspline,
    _are_collinear,
    _get_orthogonal_projection_onto_segment,
)


class DraggableArcDiagram(ArcDiagram, DraggableGraph):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_dragging_clicking_and_selecting()


    def _update_node_positions(self, nodes, cursor_position):
        # cursor_position[1] = 0. # remove y-component to remain on the line
        for node in nodes:
            x, _ = cursor_position + self._offset[self.node_artists[node]]
            self.node_positions[node] = np.array([x, self.node_positions[node][1]])


class InteractiveArcDiagram(DraggableArcDiagram, EmphasizeOnHoverGraph, AnnotateOnClickGraph, TableOnClickGraph):
    """Extends the :py:class:`ArcDiagram` class to support node placement with the mouse, emphasis of graph elements when hovering over them, and toggleable annotations.

    - Nodes can be selected and dragged around with the mouse.
    - Nodes and edges are emphasized when hovering over them.
    - Supports additional annotations that can be toggled on and off by clicking on the corresponding node or edge.
    - These annotations can also be tables.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - networkx.Graph, igraph.Graph, or graph_tool.Graph object

    node_order : list or None, default None
        The ordering of nodes (left-to-right).
        If None, the node order is optimised such that the number of edge crossings is minimal.
    above : bool, default True
        If True, edges arc above the line of nodes.
        If False, edges arc below.
    node_layout : str or dict, default 'linear'
        If `node_layout` is a string, the node positions are computed using the indicated method:

        - 'linear'    : place nodes one a horizontal line

        If `node_layout` is a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_linear_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_cmap : matplotlib color map (default 'RdGy')
        Color map used to map edge weights to edge colors. Should be diverging.
        If edge weights are strictly positive, weights are mapped to the
        left hand side of the color map with vmin=0 and vmax=np.max(weights).
        If edge weights are positive and negative, then weights are mapped
        to colors such that a weight of zero corresponds to the center of the
        color map; the boundaries are set to +/- the maximum absolute weight.
        If the graph is unweighted or the edge colors are specified explicitly,
        this parameter is ignored.
    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to :py:func:`get_arced_edge_paths`.
        Possible keyword arguments are:

        - :code:`rad` : float, default 1.
          Controls the curvature of the arcs.
        - :code:`selfloop_radius` : float, default 0.05 * np.linalg.norm(scale)
          Selfloop radius.
        - :code:`selfloop_angle` : float, default np.pi / 2
          Selfloop start angle

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().
    annotations : dict
        Mapping of nodes or edges to strings or dictionaries, the annotations.
        The visibility of the annotations can be toggled on or off by clicking on the corresponding node or edge.

        .. code-block::

           annotations = {
               0      : 'Normal node',
               1      : {s : 'Less important node', fontsize : 2},
               2      : {s : 'Very important node', fontcolor : 'red'},
               (0, 1) : 'Normal edge',
               (1, 2) : {s : 'Less important edge', fontsize : 2},
               (2, 0) : {s : 'Very important edge', fontcolor : 'red'},
           }

    annotation_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text if only the annotation string is given.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - horizontalalignment (depends on node position or edge orientation),
        - verticalalignment (depends on node position or edge orientation),
        - clip_on (default here: False),
        - backgroundcolor (default here: 'white'),
        - zorder (default here: inf),

    tables : dict node/edge : pandas dataframe
        Mapping of nodes and/or edges to pandas dataframes.
        The visibility of the tables that can toggled on or off by clicking on the corresponding node or edge.
    table_kwargs : dict
        Keyword arguments passed to matplotlib.pyplot.table.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`ArcDiagram`, :py:class:`EditableArcDiagram`

    Notes
    -----
    You must retain a reference to the plot instance!
    Otherwise, the plot instance will be garbage collected after the initial draw
    and you won't be able to move the plot elements around.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from netgraph import InteractiveArcDiagram
    >>> plt.ion()
    >>> plot_instance = InteractiveArcDiagram(my_graph_obj)
    >>> plt.show()

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._setup_emphasis()
        self._setup_annotations()
        self._setup_table_annotations()


    def _on_motion(self, event):
        DraggableArcDiagram._on_motion(self, event)
        EmphasizeOnHoverGraph._on_motion(self, event)


    def _on_release(self, event):
        if self._currently_dragging is False:
            DraggableArcDiagram._on_release(self, event)
            if self.artist_to_annotation:
                AnnotateOnClickGraph._on_release(self, event)
            if hasattr(self, 'artist_to_table'):
                TableOnClickGraph._on_release(self, event)
        else:
            DraggableArcDiagram._on_release(self, event)
            if self.artist_to_annotation:
                self._redraw_annotations(event)


    def _redraw_annotations(self, event):
        if event.inaxes == self.ax:
            for artist in self.annotated_artists:
                self._remove_annotation(artist)
                placement = self._get_annotation_placement(artist)
                self._add_annotation(artist, *placement)
            self.fig.canvas.draw()


class NascentEdge(plt.Line2D):

    def __init__(self, source, origin, rad=1., above=True):
        self.source = source
        self.origin = origin
        self.rad = rad
        self.above = above
        x, y = self._get_arc(self.origin, self.origin).T
        super().__init__(x, y, color='lightgray', linestyle='--')

    def _get_arc(self, p0, p1):
        arc_factory = ConnectionStyle.Arc3(rad=self.rad)
        path = arc_factory(p0, p1, shrinkA=0., shrinkB=0.)
        smoothed = _bspline(path.vertices, 100)
        return _lateralize(smoothed, p0, p1, self.above)

    def _update(self, x1, y1):
        x, y = self._get_arc(self.origin, (x1, y1)).T
        super().set_data(x, y)


class MutableArcDiagram(InteractiveArcDiagram, MutableGraph):
    """Extends :py:class:`InteractiveArcDiagram` to support the addition or removal of nodes and edges.

    - Double clicking on two nodes successively will create an edge between them.
    - Pressing 'insert' or '+' will add a new node to the graph.
    - Pressing 'delete' or '-' will remove selected nodes and edges.
    - Pressing '@' will reverse the direction of selected edges.

    When adding a new node, the properties of the last selected node will be used to style the node artist.
    Ditto for edges. If no node or edge has been previously selected the first created node or edge artist will be used.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - :code:`networkx.Graph`, :code:`igraph.Graph`, or :code:`graph_tool.Graph` object

    *args, **kwargs
        Parameters passed through to :py:class:`InteractiveArcDiagram`.
        See its documentation for a full list of available arguments.

    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    See also
    --------
    :py:class:`InteractiveArcDiagram`, :py:class:`MutableGraph`, :py:class:`EditableArcDiagram`

    """

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._reverse_node_artists = {artist : node for node, artist in self.node_artists.items()}
        self._reverse_edge_artists = {artist : edge for edge, artist in self.edge_artists.items()}
        self._last_selected_node_properties = self._extract_node_properties(next(iter(self.node_artists.values())))
        self._last_selected_edge_properties = self._extract_edge_properties(next(iter(self.edge_artists.values())))
        self._nascent_edge = None

        self.fig.canvas.mpl_connect('key_press_event', self._on_key_press)


    def _on_press(self, event):
        InteractiveArcDiagram._on_press(self, event)

        if event.inaxes == self.ax:
            for artist in self._clickable_artists:
                if artist.contains(event)[0]:
                    self._extract_artist_properties(artist)
                    break

            if event.dblclick:
                self._add_or_remove_nascent_edge(event)


    def _add_nascent_edge(self, node):
        nascent_edge = NascentEdge(node, self.node_positions[node], above=self.above)
        self.ax.add_artist(nascent_edge)
        return nascent_edge


    def _on_motion(self, event):
        super()._on_motion(event)

        if event.inaxes == self.ax:
            if self._nascent_edge:
                self._nascent_edge._update(event.xdata, event.ydata)
                self.fig.canvas.draw_idle()


    def _set_position_of_newly_created_node(self, x, y):
        node_positions = list(self.node_positions.values())
        if _are_collinear(node_positions):
            line = node_positions[:2]
            x, y = _get_orthogonal_projection_onto_segment((x, y), line)
        return (x, y)


class EditableArcDiagram(MutableArcDiagram, EditableGraph):
    """Extends :py:class:`InteractiveArcDiagram` to support adding, deleting, and editing graph elements interactively.

    a) Addition and removal of nodes and edges:

    - Double clicking on two nodes successively will create an edge between them.
    - Pressing 'insert' or '+' will add a new node to the graph.
    - Pressing 'delete' or '-' will remove selected nodes and edges.
    - Pressing '@' will reverse the direction of selected edges.

    b) Creation and editing of labels and annotations:

    - To create or edit a node or edge label, select the node (or edge) artist, press the 'enter' key, and type.
    - To create or edit an annotation, select the node (or edge) artist, press 'alt'+'enter', and type.
    - Terminate either action by pressing 'enter' or 'alt'+'enter' a second time.

    When adding a new node, the properties of the last selected node will be used to style the node artist.
    Ditto for edges. If no node or edge has been previously selected the first created node or edge artist will be used.

    Parameters
    ----------
    graph : various formats
        Graph object to plot. Various input formats are supported.
        In order of precedence:

        - Edge list:
          Iterable of (source, target) or (source, target, weight) tuples,
          or equivalent (E, 2) or (E, 3) ndarray, where E is the number of edges.
        - Adjacency matrix:
          Full-rank (V, V) ndarray, where V is the number of nodes/vertices.
          The absence of a connection is indicated by a zero.

          .. note:: If V <= 3, any (2, 2) or (3, 3) matrices will be interpreted as edge lists.

        - networkx.Graph, igraph.Graph, or graph_tool.Graph object

    node_order : list or None, default None
        The ordering of nodes (left-to-right).
        If None, the node order is optimised such that the number of edge crossings is minimal.
    above : bool, default True
        If True, edges arc above the line of nodes.
        If False, edges arc below.
    node_layout : str or dict, default 'linear'
        If `node_layout` is a string, the node positions are computed using the indicated method:

        - 'linear'    : place nodes one a horizontal line

        If `node_layout` is a dict, keys are nodes and values are (x, y) positions.
    node_layout : str or dict, default 'linear'
        If `node_layout` is a string, the node positions are computed using the indicated method:

        - 'linear'    : place nodes one a horizontal line

        If `node_layout` is a dict, keys are nodes and values are (x, y) positions.
    node_layout_kwargs : dict or None, default None
        Keyword arguments passed to node layout functions.
        See the documentation of the following functions for a full description of available options:

        - :py:func:`get_linear_layout`

    node_shape : str or dict, default 'o'
        Node shape.
        If the type is str, all nodes have the same shape.
        If the type is dict, maps each node to an individual string representing the shape.
        The string specification is as for matplotlib.scatter marker, i.e. one of 'so^>v<dph8'.
    node_size : float or dict, default 3.
        Node size (radius).
        If the type is float, all nodes will have the same size.
        If the type is dict, maps each node to an individual size.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_edge_width : float or dict, default 0.5
        Line width of node marker border.
        If the type is float, all nodes have the same line width.
        If the type is dict, maps each node to an individual line width.

        .. note:: Values are rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    node_color : matplotlib color specification or dict, default 'w'
        Node color.
        If the type is a string or RGBA array, all nodes have the same color.
        If the type is dict, maps each node to an individual color.
    node_edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Node edge color.
        If the type is a string or RGBA array, all nodes have the same edge color.
        If the type is dict, maps each node to an individual edge color.
    node_alpha : scalar or dict, default 1.
        Node transparency.
        If the type is a float, all nodes have the same transparency.
        If the type is dict, maps each node to an individual transparency.
    node_zorder : int or dict, default 2
        Order in which to plot the nodes.
        If the type is an int, all nodes have the same zorder.
        If the type is dict, maps each node to an individual zorder.
    node_labels : bool or dict, (default False)
        If False, the nodes are unlabelled.
        If True, the nodes are labelled with their node IDs.
        If the node labels are to be distinct from the node IDs, supply a dictionary mapping nodes to node labels.
        Only nodes in the dictionary are labelled.
    node_label_offset: float or tuple, default (0., 0.)
        A (dx, dy) tuple specifies the exact offset from the node position.
        If a single scalar delta is specified, the value is interpreted as a distance,
        and the label is placed delta away from the node position while trying to
        reduce node/label, node/edge, and label/label overlaps.
    node_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`size` (adjusted to fit into node artists if offset is (0, 0))
        - :code:`horizontalalignment` (default here: :code:`'center'`)
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: :code:`False`)
        - :code:`zorder` (default here: :code:`inf`)

    edge_width : float or dict, default 1.
        Width of edges.
        If the type is a float, all edges have the same width.
        If the type is dict, maps each edge to an individual width.

        .. note:: Value is rescaled by :py:const:`BASE_SCALE` (0.01) to be compatible with layout routines in igraph and networkx.

    edge_cmap : matplotlib color map (default 'RdGy')
        Color map used to map edge weights to edge colors. Should be diverging.
        If edge weights are strictly positive, weights are mapped to the
        left hand side of the color map with vmin=0 and vmax=np.max(weights).
        If edge weights are positive and negative, then weights are mapped
        to colors such that a weight of zero corresponds to the center of the
        color map; the boundaries are set to +/- the maximum absolute weight.
        If the graph is unweighted or the edge colors are specified explicitly,
        this parameter is ignored.
    edge_color : matplotlib color specification or dict, default :py:const:`DEFAULT_COLOR`
        Edge color.
        If the type is a string or RGBA array, all edges have the same color.
        If the type is dict, maps each edge to an individual color.
    edge_alpha : float or dict, default 1.
        The edge transparency,
        If the type is a float, all edges have the same transparency.
        If the type is dict, maps each edge to an individual transparency.
    edge_zorder : int or dict, default 1
        Order in which to plot the edges.
        If the type is an int, all edges have the same zorder.
        If the type is dict, maps each edge to an individual zorder.
        If None, the edges will be plotted in the order they appear in the 'graph' argument.
        Hint: graphs typically appear more visually pleasing if darker edges are plotted on top of lighter edges.
    arrows : bool, default False
        If True, draw edges with arrow heads.
    edge_layout_kwargs : dict, default None
        Keyword arguments passed to :py:func:`get_arced_edge_paths`.
        Possible keyword arguments are:

        - :code:`rad` : float, default 1.
          Controls the curvature of the arcs.
        - :code:`selfloop_radius` : float, default 0.05 * np.linalg.norm(scale)
          Selfloop radius.
        - :code:`selfloop_angle` : float, default np.pi / 2
          Selfloop start angle

    edge_labels : bool or dict, default False
        If False, the edges are unlabelled.
        If True, the edges are labelled with their edge IDs.
        If the edge labels are to be distinct from the edge IDs, supply a dictionary mapping edges to edge labels.
        Only edges in the dictionary are labelled.
    edge_label_position : float, default 0.5
        Relative position along the edge where the label is placed.

        - head   : 0.
        - centre : 0.5
        - tail   : 1.

    edge_label_rotate : bool, default True
        If True, edge labels are rotated such that they have the same orientation as their edge.
        If False, edge labels are not rotated; the angle of the text is parallel to the axis.
    edge_label_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text.
        For a full list of available arguments see the matplotlib documentation.
        The following default values differ from the defaults for matplotlib.text.Text:

        - :code:`horizontalalignment` (default here: :code:`'center'`),
        - :code:`verticalalignment` (default here: :code:`'center'`)
        - :code:`clip_on` (default here: code:`False`),
        - :code:`bbox` (default here: :code:`dict(color='white', pad=0)`,
        - :code:`zorder` (default here: :code:`inf`),
        - :code:`rotation` (determined by :code:`edge_label_rotate` argument)

    origin : tuple, default (0., 0.)
        The lower left hand corner of the bounding box specifying the extent of the canvas.
    scale : tuple, default (1., 1.)
        The width and height of the bounding box specifying the extent of the canvas.
    prettify : bool, default True
        If True, despine and remove ticks and tick labels.
        Set figure background to white. Set axis aspect to equal.
    ax : matplotlib.axis instance or None, default None
        Axis to plot onto; if none specified, one will be instantiated with plt.gca().
    annotations : dict
        Mapping of nodes or edges to strings or dictionaries, the annotations.
        The visibility of the annotations can be toggled on or off by clicking on the corresponding node or edge.

        .. code-block::

           annotations = {
               0      : 'Normal node',
               1      : {s : 'Less important node', fontsize : 2},
               2      : {s : 'Very important node', fontcolor : 'red'},
               (0, 1) : 'Normal edge',
               (1, 2) : {s : 'Less important edge', fontsize : 2},
               (2, 0) : {s : 'Very important edge', fontcolor : 'red'},
           }

    annotation_fontdict : dict
        Keyword arguments passed to matplotlib.text.Text if only the annotation string is given.
        For a full list of available arguments see the matplotlib documentation.
    tables : dict node/edge : pandas dataframe
        Mapping of nodes and/or edges to pandas dataframes.
        The visibility of the tables that can toggled on or off by clicking on the corresponding node or edge.
    table_kwargs : dict
        Keyword arguments passed to matplotlib.pyplot.table.


    Attributes
    ----------
    node_artists : dict
        Mapping of node IDs to matplotlib PathPatch artists.
    edge_artists : dict
        Mapping of edge IDs to matplotlib PathPatch artists.
    node_label_artists : dict
        Mapping of node IDs to matplotlib text objects (if applicable).
    edge_label_artists : dict
        Mapping of edge IDs to matplotlib text objects (if applicable).
    node_positions : dict node : (x, y) tuple
        Mapping of node IDs to node positions.

    Notes
    -----
    You must retain a reference to the plot instance!
    Otherwise, the plot instance will be garbage collected after the initial draw
    and you won't be able to move the plot elements around.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from netgraph import InteractiveArcDiagram
    >>> plt.ion()
    >>> plot_instance = EditableArcDiagram(my_graph_obj)
    >>> plt.show()

    See also
    --------
    :py:class:`InteractiveArcDiagram`, :py:class:`EditableGraph`

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # initiate node and edge label data structures if they don't exist
        if not hasattr(self, 'node_label_artists'):
            node_labels = {node : '' for node in self.nodes}
            self.node_label_fontdict = self._initialize_node_label_fontdict(
                kwargs.get('node_label_fontdict'), node_labels, kwargs.get('node_label_offset', (0., 0.)))
            self.node_label_offset, self._recompute_node_label_offsets =\
                self._initialize_node_label_offset(node_labels, kwargs.get('node_label_offset', (0., 0.)))
            if self._recompute_node_label_offsets:
                self._update_node_label_offsets()
            self.node_label_artists = dict()
            self.draw_node_labels(node_labels, self.node_label_fontdict)

        if not hasattr(self, 'edge_label_artists'):
            edge_labels = {edge : '' for edge in self.edges}
            self.edge_label_fontdict = self._initialize_edge_label_fontdict(kwargs.get('edge_label_fontdict'))
            self.edge_label_position = kwargs.get('edge_label_position', 0.5)
            self.edge_label_rotate = kwargs.get('edge_label_rotate', True)
            self.edge_label_artists = dict()
            self.draw_edge_labels(edge_labels, self.edge_label_position,
                                  self.edge_label_rotate, self.edge_label_fontdict)

        self._currently_writing_labels = False
        self._currently_writing_annotations = False



"""
This module provides various classes for Transfer Entropy-based feature selection analysis.

Classes:
    TEestimator: support class for handling different Transfer Entropy estimators.
    TE: Transfer Entropy class.
"""

import copy
from enum import Enum
import numpy as np
from causalflow.selection_methods.SelectionMethod import SelectionMethod, CTest, _suppress_stdout
from idtxl.multivariate_te import MultivariateTE
from idtxl.bivariate_mi import BivariateMI
from idtxl.data import Data
from causalflow.CPrinter import CP
from scipy.stats import shapiro, kstest
import importlib


class TEestimator(Enum):
    """TEestimator Enumerator."""
    
    Auto = 'Auto'
    Gaussian = 'JidtGaussianCMI'
    Kraskov = 'JidtKraskovCMI'
    OpenCLKraskov = 'OpenCLKraskovCMI'


class TE(SelectionMethod):
    """Feature selection method based on Trasfer Entropy analysis."""
    
    def __init__(self, estimator: TEestimator):
        """
        Class contructor.

        Args:
            estimator (TEestimator): Gaussian/Kraskov.
        """
        super().__init__(CTest.TE)
        self.estimator = estimator
        
    @property
    def isOpenCLinstalled(self) -> bool:
        """
        Check whether the pyopencl pkg is installed.

        Returns:
            bool: True if pyopencl is installed.
        """
        try:
            importlib.import_module('pyopencl')
            return True
        except ImportError:
            return False
        
    def _select_estimator(self):
        """Select the TE estimator."""
        CP.info("\n##")
        CP.info("## TE Estimator selection")
        CP.info("##")

        isGaussian = True

        for f in self.data.features:
            # Perform Shapiro-Wilk test
            shapiro_stat, shapiro_p_value = shapiro(self.data.d[f])
            # Perform Kolmogorov-Smirnov test
            ks_stat, ks_p_value = kstest(self.data.d[f], 'norm')
                
            # Print results
            CP.debug("\n")
            CP.debug(f"Feature '{f}':")
            CP.debug(f"\t- Shapiro-Wilk test: val={round(shapiro_stat, 2)}, p-val={round(shapiro_p_value, 2)}")
            CP.debug(f"\t- Kolmogorov-Smirnov test: val={round(ks_stat, 2)}, p-val={round(ks_p_value, 2)}")
                
            # Check if p-values are less than significance level (e.g., 0.05) for normality
            if shapiro_p_value < 0.05 or ks_p_value < 0.05:
                CP.debug("\tNot normally distributed")
                isGaussian = False
                # break
            else:
                CP.debug("\tNormally distributed")
                
        if isGaussian:
            self.estimator = TEestimator.Gaussian
        else:
            self.estimator = TEestimator.OpenCLKraskov if self.isOpenCLinstalled else TEestimator.Kraskov
        CP.info("\n## TE Estimator: " + self.estimator.value)


    def compute_dependencies(self):
        """
        Compute list of dependencies for each target by transfer entropy analysis.

        Returns:
            (DAG): dependency dag.
        """
        if self.estimator is TEestimator.Auto: self._select_estimator()

        multi_network_analysis = MultivariateTE()
        bi_network_analysis = BivariateMI()
        cross_settings = {'cmi_estimator': self.estimator.value,
                    'max_lag_sources': self.max_lag,
                    'min_lag_sources': self.min_lag,
                    'max_lag_target': self.max_lag,
                    'min_lag_target': self.min_lag,
                    'alpha_max_stats': self.alpha,
                    'alpha_min_stats': self.alpha,
                    'alpha_omnibus': self.alpha,
                    'alpha_max_seq': self.alpha,
                    'verbose': False}
        autodep_settings = copy.deepcopy(cross_settings)
        if self.min_lag == 0:
            autodep_settings['min_lag_sources'] = 1
        
        CP.info("\n##")
        CP.info("## " + self.name + " analysis")
        CP.info("##")
        for target in self.data.features:
            CP.info("\n## Target variable: " + target)
            with _suppress_stdout():
                t = self.data.features.index(target)
                
                # Check auto-dependency
                tmp_d = np.c_[self.data.d.values[:, t], self.data.d.values[:, t]]
                data = Data(tmp_d, dim_order='sp') # sp = samples(row) x processes(col)
                res_auto = bi_network_analysis.analyse_single_target(settings = autodep_settings, data = data, target = 0, sources = 1)
                
                # Check cross-dependencies
                data = Data(self.data.d.values, dim_order='sp') # sp = samples(row) x processes(col)
                res_cross = multi_network_analysis.analyse_single_target(settings = cross_settings, data = data, target = t)
            
            # Auto-dependency handling
            auto_lag = [s[1] for s in res_auto._single_target[0]['selected_vars_sources']]
            auto_score = res_auto._single_target[0]['selected_sources_mi']
            auto_pval = res_auto._single_target[0]['selected_sources_pval']
            if auto_score is not None:
                for score, pval, lag in zip(auto_score, auto_pval, auto_lag):
                    self._add_dependency(self.data.features[t], self.data.features[t], score, pval, lag)
            
            # Cross-dependencies handling    
            sel_sources = [s[0] for s in res_cross._single_target[t]['selected_vars_sources']]
            if sel_sources:
                sel_sources_lag = [s[1] for s in res_cross._single_target[t]['selected_vars_sources']]
                sel_sources_score = res_cross._single_target[t]['selected_sources_te']
                sel_sources_pval = res_cross._single_target[t]['selected_sources_pval']
                for s, score, pval, lag in zip(sel_sources, sel_sources_score, sel_sources_pval, sel_sources_lag):
                    self._add_dependency(self.data.features[t], self.data.features[s], score, pval, lag)
            
            if auto_score is None and not sel_sources:
                CP.info("\tno sources selected")

        return self.result


"""
This module provides subsampling methods for data preprocessing.

Classes:
    SSMode: An enumerator containing all the supported subsampling methods.
    SubsamplingMethod: A class for implementing various subsampling techniques.
"""

from abc import ABC, abstractmethod
from enum import Enum
import pandas as pd


class SSMode(Enum):
    """Enumerator containing all the supported subsampling methods."""

    WSDynamic = 'Dynamic-size moving window'
    WSStatic = 'Static-size moving window'
    WSFFTStatic = 'FFT static-size moving window'
    Static = 'Static'
    

class SubsamplingMethod(ABC):
    """SubsamplingMethod abstract class."""
    
    def __init__(self, ssmode: SSMode):
        """
        Class constructor.

        Args:
            ssmode (SSMore): Subsampling method.
        """
        self.ssmode = ssmode
        self.df = None

    
    def initialise(self, dataframe: pd.DataFrame):
        """
        Initialise class by setting the dataframe to subsample.

        Args:
            dataframe (pd.DataFrame): Pandas DataFrame to subsample.
        """
        self.df = dataframe


    @abstractmethod
    def run(self):
        """Run subsampler."""
        pass



#!/usr/bin/env python
# coding: utf-8
"""
Netgraph utility functions.
"""

import numpy as np
import matplotlib as mpl

from numpy.linalg import matrix_rank
from scipy.interpolate import BSpline
from matplotlib.path import Path
from scipy.spatial import cKDTree
from scipy.interpolate import interp1d
from scipy.optimize import minimize_scalar


def _save_cast_float_to_int(num):
    """Cast number to an integer if it is close to an integer."""
    if isinstance(num, (float, int)) and np.isclose(num, int(num)):
        return int(num)
    return num


def _save_cast_str_to_int(string):
    """Cast string to an integer if possible."""
    try:
        return int(string)
    except:
        return string


def _get_unique_nodes(edges):
    """
    Parameters
    ----------
    edges: list of tuple
        Edge list of the graph.

    Returns
    -------
    nodes: list
        List of unique nodes.

    Notes
    -----
    We cannot use numpy.unique, as it promotes nodes to numpy.float/numpy.int/numpy.str,
    and breaks for nodes that have a more complicated type such as a tuple.

    """
    return list(set(_flatten([(edge[0], edge[1]) for edge in edges])))


def _flatten(nested_list):
    """Flatten a nested list."""
    return [item for sublist in nested_list for item in sublist]


def _edge_list_to_adjacency_matrix(edges, edge_weights=None, unique_nodes=None):
    """Convert an edge list representation of a graph into the corresponding full rank adjacency matrix.

    Parameters
    ----------
    edges : list of tuple
        List of edges; each edge is identified by a (v1, v2) node tuple.
    edge_weights : list of int or float, optional
        List of corresponding edge weights.
    unique_nodes : list
        List of unique nodes. Required if any node is unconnected.

    Returns
    -------
    adjacency_matrix : numpy.array
        The full rank adjacency/weight matrix.

    """

    sources = [s for (s, _) in edges]
    targets = [t for (_, t) in edges]
    if edge_weights:
        weights = [edge_weights[edge] for edge in edges]
    else:
        weights = np.ones((len(edges)))

    if unique_nodes is None:
        # map nodes to consecutive integers
        nodes = sources + targets
        unique_nodes = set(nodes)

    indices = range(len(unique_nodes))
    node_to_idx = dict(zip(unique_nodes, indices))

    source_indices = [node_to_idx[source] for source in sources]
    target_indices = [node_to_idx[target] for target in targets]

    total_nodes = len(unique_nodes)
    adjacency_matrix = np.zeros((total_nodes, total_nodes))
    adjacency_matrix[source_indices, target_indices] = weights

    return adjacency_matrix


def _edge_list_to_adjacency_list(edges, directed=True):
    """Convert an edge list representation of a unweighted graph into the corresponding adjacency list representation.

    Parameters
    ----------
    edges : list of tuple
        List of edges; each edge is identified by a (node, node) tuple.
    directed : bool, default True
        Indicates if the graph is directed.

    Returns
    -------
    adjacency : dict node : set of nodes
        Dictionary mapping nodes to their set of connected neighbours.

    """
    if not directed:
        edges = edges + [(target, source) for (source, target) in edges] # forces copy

    adjacency = dict()
    for source, target in edges:
        if source in adjacency:
            adjacency[source] |= set([target])
        else:
            adjacency[source] = set([target])
    return adjacency


def _get_subgraph(edges, nodes):
    """Induce the subgraph using the specified nodes.

    Parameters
    ----------
    edges : list of tuple
        List of edges; each edge is identified by a (v1, v2) node tuple.
    nodes : list
        List of nodes.

    Returns
    -------
    subgraph_edges : list of tuple
        List of edges present in the subgraph.

    """

    return [(source, target) for source, target in edges \
            if (source in nodes) and (target in nodes)]


def _bspline(cv, n=100, degree=5, periodic=False):
    """Calculate n samples on a bspline.

    Parameters
    ----------
    cv : numpy.array
        Array of (x, y) control vertices.
    n : int
        Number of samples to return.
    degree : int
        Curve degree
    periodic : bool, default True
        If True, the curve is closed.

    Returns
    -------
    numpy.array
        Array of (x, y) spline vertices.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/35007804/2912349

    """

    cv = np.asarray(cv)
    count = cv.shape[0]

    # Closed curve
    if periodic:
        kv = np.arange(-degree,count+degree+1)
        factor, fraction = divmod(count+degree+1, count)
        cv = np.roll(np.concatenate((cv,) * factor + (cv[:fraction],)),-1,axis=0)
        degree = np.clip(degree,1,degree)

    # Opened curve
    else:
        degree = np.clip(degree,1,count-1)
        kv = np.clip(np.arange(count+degree+1)-degree,0,count-degree)

    # Return samples
    max_param = count - (degree * (1-periodic))
    spl = BSpline(kv, cv, degree)
    return spl(np.linspace(0,max_param,n))


def _get_angle(dx, dy, radians=True):
    """Angle of a vector in 2D."""
    angle = np.arctan2(dy, dx)
    if radians:
        return angle
    else:
        return angle * 360 / (2.0 * np.pi)


def _get_interior_angle_between(v1, v2, radians=True):
    """Returns the interior angle between vectors v1 and v2.

    Parameters
    ----------
    v1, v2 : numpy.array
        The vectors in question.
    radians : bool, default True
        If True, return the angle in radians (otherwise it is in degrees).

    Returns
    -------
    angle : float
        The interior angle between two vectors.

    Examples
    --------
    >>> angle_between((1, 0, 0), (0, 1, 0))
    1.5707963267948966
    >>> angle_between((1, 0, 0), (1, 0, 0))
    0.0
    >>> angle_between((1, 0, 0), (-1, 0, 0))
    3.141592653589793

    Notes
    -----
    Adapted from https://stackoverflow.com/a/13849249/2912349

    See also
    --------
    _get_signed_angle_between

    """

    v1_u = _get_unit_vector(v1)
    v2_u = _get_unit_vector(v2)
    angle = np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))
    if radians:
        return angle
    else:
        return angle * 360 / (2 * np.pi)


def _get_unit_vector(vector):
    """Returns the unit vector of the vector."""
    return vector / np.linalg.norm(vector)


def _get_signed_angle_between(v1, v2, radians=True):
    """Returns the signed angle between vectors v1 and v2.

    Parameters
    ----------
    v1, v2 : numpy.array
        The vectors in question.
    radians : bool, default True
        If True, return the angle in radians (otherwise it is in degrees).

    Returns
    -------
    angle : float
        The signed angle between two vectors.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/16544330/2912349

    See also
    --------
    _get_interior_angle_between

    """

    x1, y1 = v1
    x2, y2 = v2
    dot = x1*x2 + y1*y2
    det = x1*y2 - y1*x2
    angle = np.arctan2(det, dot)
    if radians:
        return angle
    else:
        return angle * 360 / (2 * np.pi)


def _get_n_points_on_a_circle(xy, radius, n, start_angle=0):
    """Determine the positions of n evenly spaced points on a circle with a given (x, y) origin and radius.

    Parameters
    ----------
    xy : tuple of float
        The origin of the circle.
    radius : float
        The radius of the circle.
    n : int
        The number of points returned.
    start_angle : float, default 0
        The angle at which the first point is placed in radians.

    Returns
    -------
    positions ; numpy.array
        Array of n (x, y) coordinates.

    """

    angles = np.linspace(0, 2*np.pi, n + 1)[:-1]
    angles = (angles + start_angle) % (2*np.pi)
    positions = np.array([_get_point_on_a_circle(xy, radius, angle) for angle in angles])
    return positions


def _get_point_on_a_circle(origin, radius, angle):
    """Compute the (x, y) coordinate of a point at a specified angle on a circle given by its (x, y) origin and radius."""

    x0, y0 = origin
    x = x0 + radius * np.cos(angle)
    y = y0 + radius * np.sin(angle)
    return np.array([x, y])


def _get_parallel_line(path, delta):
    """Compute a parallel to a given path with an offset delta.

    Parameters
    ----------
    path : numpy.array
        Array of (x, y) path coordinates.
    delta : float
        Offset from the path.

    Returns
    -------
    path : numpy.array
        Array of (x, y) coordinates corresponding to the parallel path.

    """

    if not _are_collinear(path):
        # initialise output
        orthogonal_unit_vector = np.zeros_like(path)

        tangents = path[2:] - path[:-2] # using the central difference approximation
        orthogonal_unit_vector[1:-1] = _get_orthogonal_unit_vector(tangents)

        # handle start and end points
        orthogonal_unit_vector[ 0] = _get_orthogonal_unit_vector(np.atleast_2d([path[ 1] - path[ 0]]))
        orthogonal_unit_vector[-1] = _get_orthogonal_unit_vector(np.atleast_2d([path[-1] - path[-2]]))

    else:
        orthogonal_unit_vector = _get_orthogonal_unit_vector(np.atleast_2d([path[-1] - path[0]]))

    return path + delta * orthogonal_unit_vector


def _get_orthogonal_unit_vector(v):
    """Determine the orthogonal unit vector to a given vector.

    Parameters
    ----------
    v : numpy.array
        The input vector.

    Returns
    -------
    w : numpy.array
        The output vector.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/16890776/2912349

    """
    if not np.all(np.isclose(v, 0)):
        v = v / np.linalg.norm(v, axis=-1)[:, None] # unit vector
        w = np.c_[-v[:,1], v[:,0]]                  # orthogonal vector
        w = w / np.linalg.norm(w, axis=-1)[:, None] # orthogonal unit vector
        return w
    else:
        return v


def _shorten_spline_by(path, distance):
    """Cut path off at the end by `distance`.

    Parameters
    ----------
    path : numpy.array
        Array of (x, y) path coordinates.
    distance : float
        Distance travelled along the path to remove from the end of the path.

    Returns
    -------
    numpy.array
        Array of (x, y) coordinates of the shortened path.

    """

    if distance > 0:
        distance_to_end = np.linalg.norm(path - path[-1], axis=1)
        is_valid = (distance_to_end - distance) >= 0
        if np.any(is_valid):
            idx = np.where(is_valid)[0][-1] # i.e. the last valid point
        else:
            idx = 0

        # We could truncate the  path using `path[:idx+1]` and return here.
        # However, if the path is not densely sampled, the error will be large.
        # Therefor, we compute a point that is on the line from the last valid point to
        # the end point, and append it to the truncated path.
        vector = path[idx] - path[-1]
        unit_vector = vector / np.linalg.norm(vector)
        new_end_point = path[-1] + distance * unit_vector

        return np.concatenate([path[:idx+1], new_end_point[None, :]], axis=0)

    else:
        return path


def _get_point_along_spline(spline, fraction):
    """Determine the point coordinates at a given fraction of the spline.

    Parameters
    ----------
    spline : numpy.array
        (N points, 2) array of (x, y) spline coordinates.
    fraction : float
        Fraction of the spline. Has to be a value between 0 and 1.

    Returns
    -------
    point : 2-tuple of floats
        The (x, y) point coordinates.

    """

    assert 0 <= fraction <= 1, "Fraction has to be a value between 0 and 1."
    deltas = np.diff(spline, axis=0)
    successive_distances = np.sqrt(np.sum(deltas**2, axis=1))
    cumulative_sum = np.cumsum(successive_distances)
    desired_length = cumulative_sum[-1] * fraction
    idx = np.where(cumulative_sum >= desired_length)[0][0] # upper bound
    overhang = cumulative_sum[idx] - desired_length
    x, y = spline[idx+1] - overhang/successive_distances[idx] * deltas[idx]
    return x, y


def _resample_spline(spline, total_samples=100):
    """Resample a spline using the given number of points.

    Parameters
    ----------
    spline : numpy.array
        (N points, 2) array of (x, y) spline coordinates.
    total_samples : int
        The number of evenly spaces points after re-sampling.

    Returns
    -------
    resampled : numpy.array
        (M samples, 2) array of (x, y) resampled spline coordinates.

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/52020098/2912349
    """
    distance = np.cumsum(np.sqrt(np.sum(np.diff(spline, axis=0)**2, axis=1)))
    distance = np.insert(distance, 0, 0)/distance[-1]
    interpolator = interp1d(distance, spline, kind='slinear', axis=0)
    return interpolator(np.linspace(0, 1, total_samples))


def _get_tangent_at_point(spline, fraction):
    """Compute the tangent to a spline at a given fraction of the spline.

    Parameters
    ----------
    spline : numpy.array
        (N points, 2) array of (x, y) spline coordinates.
    fraction : float
        Fraction of the spline. Has to be a value between 0 and 1.

    Returns
    -------
    tangent : numpy.array
        The (dx, dy) tangent.

    """

    assert 0 <= fraction <= 1, "Fraction has to be a value between 0 and 1."
    deltas = np.diff(spline, axis=0)
    successive_distances = np.sqrt(np.sum(deltas**2, axis=1))
    cumulative_sum = np.cumsum(successive_distances)
    desired_length = cumulative_sum[-1] * fraction
    idx = np.where(cumulative_sum >= desired_length)[0][0] # upper bound
    return deltas[idx]


def _get_orthogonal_projection_onto_segment(point, segment):
    """Given a segment defined by points P1 and P2, determine the orthogonal projection of a point P3 onto said segment.

    Parameters
    ----------
    segment : tuple of (float x, float y) tuples
        The segment defined by (P1, P2).
    point : (float x, float y) tuple
        The point P3 to be projected onto the segment.

    Returns
    -------
    projected_point : (float x, float y)
        The coordinates of the projected point.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/61343727/2912349

    """

    p1, p2 = segment

    segment_length = np.sum((p1-p2)**2)

    # The line extending the segment is parameterized as p1 + t (p2 - p1).
    # The projection falls where t = [(point-p1) . (p2-p1)] / |p2-p1|^2
    t = np.sum((point - p1) * (p2 - p1)) / segment_length

    # # Project onto line segment between p1 and p2 or closest point of the line segment.
    # t = max(0, t)

    return p1 + t * (p2 - p1)


def _find_renderer(fig):
    """
    Return the renderer for a given matplotlib figure.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/22689498/2912349

    """

    if hasattr(fig.canvas, "get_renderer"):
        # Some backends, such as TkAgg, have the get_renderer method, which
        # makes this easy.
        renderer = fig.canvas.get_renderer()
    else:
        # Other backends do not have the get_renderer method, so we have a work
        # around to find the renderer. Print the figure to a temporary file
        # object, and then grab the renderer that was used.
        # (I stole this trick from the matplotlib backend_bases.py
        # print_figure() method.)
        import io
        fig.canvas.print_pdf(io.BytesIO())
        renderer = fig._cachedRenderer
    return(renderer)


def _make_pretty(ax):
    """Remove the figure frame, x- and y-ticks, and set the aspect to equal."""
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_aspect('equal')
    ax.get_figure().set_facecolor('w')
    # ax.set_frame_on(False) # also removes background
    for spine in ax.spines.values():
        spine.set_visible(False)
    ax.get_figure().canvas.draw()


def _rank(vec):
    """Compute the rank for each value in a vector."""
    tmp = np.argsort(vec)
    ranks = np.empty_like(vec)
    ranks[tmp] = np.arange(len(vec))
    return ranks


def _invert_dict(mydict):
    """Invert a dictionary such that values map to keys."""
    inverse = dict()
    for key, value in mydict.items():
        inverse.setdefault(value, set()).add(key)
    return inverse


def _get_connected_components(adjacency_list):
    """
    Get the connected components given a graph in adjacency list format.

    Parameters
    ---------
    adjacency_list : dict node ID : set of node IDs
        Adjacency list, i.e. a mapping from each node to its neighbours.

    Returns
    -------
    components : list of sets of node IDs
        The unconnected components of the graph.

    """

    components = []
    not_visited = set(list(adjacency_list.keys()))
    while not_visited: # i.e. while stack is non-empty (empty set is interpreted as `False`)
        start = not_visited.pop()
        component = _bfs(adjacency_list, start)
        components.append(component)

        #  remove nodes that are in the component that we just found
        for node in component:
            try:
                not_visited.remove(node)
            except KeyError:
                # KeyErrors occur when we try to remove
                # 1) the start node (which we already popped), or
                # 2) leaf nodes, i.e. nodes with no outgoing edges
                pass

    # Often, we are only interested in the largest component,
    # hence we return the list of components sorted by size, largest first.
    components = sorted(components, key=len, reverse=True)

    return components


def _dfs(adjacency_list, start, visited=None):
    """Depth first search on a given adjacency list.

    Parameters
    ----------
    adjacency_list : dict node ID : set of node IDs
        Adjacency list, i.e. a mapping from each node to its neighbours.

    start : node ID
        The starting node.

    visited : set of node IDs or None, default None
        Previously visited nodes.

    Returns
    -------
    visited : set of node IDs or None, default None
        Previously and newly visited nodes.

    """

    if visited is None:
        visited = set()
    visited.add(start)
    for node in adjacency_list[start] - visited:
        if node in adjacency_list:
            _dfs(adjacency_list, node, visited)
        else: # otherwise no outgoing edge
            visited.add(node)
    return visited


def _bfs(adjacency_list, start):
    """A fast BFS node generator.

    Parameters
    ----------
    adjacency_list : dict node ID : set of node IDs
        Adjacency list, i.e. a mapping from each node to its neighbours.
    start : node ID
        The starting node.

    Returns
    -------
    visited : set of node IDs or None, default None
        All reachable nodes from source.

    Notes
    -----
    Adapted from networkx.algorithms.components._plain_bfs
    """

    visited = {start}
    next_level = [start]
    while next_level:
        this_level = next_level
        next_level = []
        for node in this_level:
            for neighbour in adjacency_list[node]:
                if neighbour not in visited:
                    visited.add(neighbour)
                    next_level.append(neighbour)
            if len(visited) == len(adjacency_list):
                return visited
    return visited


def _get_gradient_and_intercept(p1, p2):
    x1, y1 = p1
    x2, y2 = p2
    if x1 != x2:
        gradient = (y2 - y1) / (x2 - x1)
        intercept = (x2 * y1 - x1 * y2) / (x2 - x1)
    elif (x1 == x2) and (y1 != y2):
        gradient = np.sign(y2 - y1) * np.inf
        intercept = x1
    else:
        gradient, intercept = np.nan, np.nan
    return gradient, intercept


def _is_above_line(points, gradient, intercept):
    """Returns true for points above a given line.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/45769740/2912349

    """
    a = np.array([0, intercept])
    b = np.array([1, intercept + gradient])
    return np.cross(points-a, b-a) < 0


def _reflect_across_line(points, gradient, intercept):
    """Reflect a set of points across a given line.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/3307181/2912349

    """
    x0, y0 = points.T
    if not np.isinf(gradient):
        d = (x0 + (y0 - intercept) * gradient) / (1 + gradient**2)
        x1 = 2 * d - x0
        y1 = 2 * d * gradient - y0 + 2 * intercept
        return np.c_[x1, y1]
    else:
        x1 = x0 + 2 * (intercept - x0)
        return np.c_[x1, y0]


def _are_collinear(points, tol=None):
    "Test if the given points are collinear."
    if np.any(np.isnan(points)):
        raise ValueError("`points` may not contain `NaN` values!")
    elif np.any(np.isinf(points)):
        raise ValueError("`points` may not contain `inf` values!")
    else:
        points = np.array(points)
        points -= points.mean(axis=0)[np.newaxis, :]
        rank = matrix_rank(points, tol=tol)
        return rank == 1


def _convert_polar_to_cartesian_coordinates(rho, phi):
    # Adapted from https://stackoverflow.com/a/26757297/2912349
    x = rho * np.cos(phi)
    y = rho * np.sin(phi)
    return(x, y)


def _normalize_numeric_argument(numeric_or_dict, dict_keys, variable_name, allow_none=False):
    if allow_none:
        allowed_types = (int, float, type(None))
    else:
        allowed_types = (int, float)
    if isinstance(numeric_or_dict, allowed_types):
        return {key : numeric_or_dict for key in dict_keys}
    elif isinstance(numeric_or_dict, dict):
        _check_completeness(numeric_or_dict, dict_keys, variable_name)
        _check_types(numeric_or_dict.values(), allowed_types, variable_name)
        return numeric_or_dict
    else:
        msg = f"The type of {variable_name} has to be either a int, float, or a dict."
        msg += f"\nThe current type is {type(numeric_or_dict)}."
        raise TypeError(msg)


def _check_completeness(given_set, desired_set, variable_name):
    # ensure that iterables are sets
    # TODO: check that iterables can safely be converted to sets (unlike dict keys)
    given_set = set(given_set)
    desired_set = set(desired_set)

    complete = given_set.issuperset(desired_set)
    if not complete:
        missing = desired_set - given_set
        msg = f"{variable_name} is incomplete. The following elements are missing:"
        for item in missing:
            if isinstance(item, str):
                msg += f"\n\'{item}\'"
            else:
                msg += f"\n{item}"
        raise ValueError(msg)


def _check_types(items, types, variable_name):
    for item in items:
        if not isinstance(item, types):
            msg = f"Item {item} in {variable_name} is of the wrong type."
            msg += f"\nExpected type: {types}"
            msg += f"\nActual type: {type(item)}"
            raise TypeError(msg)


def _normalize_string_argument(str_or_dict, dict_keys, variable_name):
    if isinstance(str_or_dict, str):
        return {key : str_or_dict for key in dict_keys}
    elif isinstance(str_or_dict, dict):
        _check_completeness(set(str_or_dict), dict_keys, variable_name)
        _check_types(str_or_dict.values(), str, variable_name)
        return str_or_dict
    else:
        msg = f"The type of {variable_name} has to be either a str or a dict."
        msg += f"The current type is {type(str_or_dict)}."
        raise TypeError(msg)


def _normalize_shape_argument(str_path_or_dict, dict_keys, variable_name):
    if isinstance(str_path_or_dict, str):
        return {key : str_path_or_dict for key in dict_keys}
    elif isinstance(str_path_or_dict, Path):
        return {key : str_path_or_dict for key in dict_keys}
    elif isinstance(str_path_or_dict, dict):
        _check_completeness(set(str_path_or_dict), dict_keys, variable_name)
        _check_types(str_path_or_dict.values(), (str, Path), variable_name)
        return str_path_or_dict
    else:
        msg = f"The type of {variable_name} has to be either a str, matplotlib.path.Path or a dict."
        msg += f"The current type is {type(str_path_or_dict)}."
        raise TypeError(msg)


def _normalize_color_argument(color_or_dict, dict_keys, variable_name):
    if mpl.colors.is_color_like(color_or_dict):
        return {key : color_or_dict for key in dict_keys}
    elif color_or_dict is None:
        return {key : color_or_dict for key in dict_keys}
    elif isinstance(color_or_dict, dict):
        _check_completeness(set(color_or_dict), dict_keys, variable_name)
        # TODO: assert that each element is a valid color
        return color_or_dict
    else:
        msg = f"The type of {variable_name} has to be either a valid matplotlib color specification or a dict."
        raise TypeError(msg)


def _rescale_dict_values(mydict, scalar):
    return {key: value * scalar for (key, value) in mydict.items()}


# # Variant no 1: use force directed layout to determine a suitable node label placements
# # pros : labels repel each other
# # cons : does not work very well; the optimum placement can still result in a collision
# def _get_optimal_offsets(self, anchors, offsets, avoid, total_iterations=5):
#     # Compute the net repulsion exerted on each label by nodes, edges and other labels.
#     # Place the label in the direction of net repulsion at the desired distance from the corresponding node (anchor).
#     # TODO Test if gradually stepping in the direction of net repulsion improves results.
#     for ii in range(total_iterations):
#         repulsion = self._get_repulsion(anchors + offsets, avoid)
#         directions = repulsion / np.linalg.norm(repulsion, axis=-1)[:, np.newaxis]
#         offsets = np.linalg.norm(offsets, axis=-1)[:, np.newaxis] * directions
#     return offsets


# def _get_repulsion(self, mobile, fixed, minimum_distance=0.01):
#     combined = np.concatenate([mobile, fixed], axis=0)

#     delta = mobile[np.newaxis, :, :] - combined[:, np.newaxis, :]
#     distance = np.linalg.norm(delta, axis=-1)
#     direction = delta / distance[..., None] # i.e. the unit vector

#     # 1. We clip the distance as we want to reduce overlaps with
#     # all nearby plot elements, not just the one that overlaps the
#     # most.
#     # 2. We only care about interactions with nearby objects, so
#     # we heavily penalise repulsion from far away items by using a
#     # exponent.
#     magnitude = 1. / np.clip(distance, minimum_distance, np.inf)**6
#     repulsion = direction * magnitude[..., None]

#     for ii in range(repulsion.shape[-1]):
#         np.fill_diagonal(repulsion[:, :, ii], 0)

#     return np.sum(repulsion, axis=0)


# Variant no 2:
# pros : straightforward optimisation; works very well
# cons : labels can still collide with each other
def _get_optimal_offsets(anchors, offsets, avoid, total_queries_per_point=360):
    """Find the location at a specified distance (`offset`) away from a
    given other location (`anchor`) that is furthest away from all
    locations in `avoid`.
    """
    tree = cKDTree(avoid)
    output = np.zeros((len(offsets), 2))
    for ii, (anchor, offset) in enumerate(zip(anchors, offsets)):
        x = _get_n_points_on_a_circle(anchor, np.linalg.norm(offset), total_queries_per_point)
        # distances, _ = tree.query(x, 1) # can result in many ties; first element is arbitrarily chosen
        # output[ii] = x[np.argmax(distances)]
        distances, _ = tree.query(x, 2)
        output[ii] = x[np.argmax(np.sum(distances, axis=1))]
    return output - anchors


def _print_progress_bar(iteration, total_iterations, prefix='', suffix='', length=100):
    """Call in a loop to create terminal progress bar.

    Parameters
    ----------
    iteration: int
        The current iteration.
    total_iterations: int
        The total number of iterations.
    prefix: str (default '')
        The prefix.
    suffix: str (default '')
        The suffix.
    length: int (default 100)
        The character length of bar.

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/34325723/2912349

    """
    fraction_complete = iteration / total_iterations
    bar = '█' * int(length * fraction_complete) + '-' * int(length * (1 - fraction_complete))
    print(f'\r{prefix} |{bar}| {100 * fraction_complete:.1f}% {suffix}', end="\r")
    if iteration == total_iterations:
        print()


def _get_total_pixels(fig):
    w, h = fig.get_size_inches()
    dpi = fig.get_dpi()
    total_pixels = w * h * dpi**2
    return total_pixels


def _simplify_multigraph(edges):
    return list({(source, target) for source, target, _ in edges})


def _map_multigraph_edges_to_ids(edges):
    output = dict()
    for source, target, eid in edges:
        output[(source, target)] = output.get((source, target), list()) + [eid]
    return output


def _get_radius(path, center=np.zeros((2))):
    """Adapted from https://stackoverflow.com/a/76064783/2912349"""
    deltas = path.vertices - center[np.newaxis, :]
    distances = np.linalg.norm(deltas, axis=-1)
    upper_bound = np.max(distances)

    def func(r):
        circle_path = Path.circle(center, radius=r)
        if circle_path.contains_path(path):
            return r
        else:
            return upper_bound

    result = minimize_scalar(func, bounds=(0, upper_bound))
    return result.x



"""
This module provides the FPCMCI class.

Classes:
    FPCMCI: class containing the FPCMCI causal discovery algorithm.
"""

import copy
import pickle
from tigramite.independence_tests.independence_tests_base import CondIndTest
from causalflow.graph.DAG import DAG
from causalflow.selection_methods.SelectionMethod import SelectionMethod
from causalflow.CPrinter import CPLevel, CP
from causalflow.basics.constants import *
from causalflow.causal_discovery.support.myPCMCI import myPCMCI
from causalflow.preprocessing.data import Data 
from causalflow.causal_discovery.CausalDiscoveryMethod import CausalDiscoveryMethod 


class FPCMCI(CausalDiscoveryMethod):
    """F-PCMCI causal discovery method."""

    def __init__(self, 
                 data: Data, 
                 min_lag, max_lag, 
                 sel_method: SelectionMethod, val_condtest: CondIndTest, 
                 verbosity: CPLevel, 
                 f_alpha = 0.05, 
                 alpha = 0.05, 
                 resfolder = None,
                 neglect_only_autodep = False,
                 clean_cls = True):
        """
        Class contructor.

        Args:
            data (Data): data to analyse.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            sel_method (SelectionMethod): selection method.
            val_condtest (CondIndTest): validation method.
            verbosity (CPLevel): verbosity level.
            f_alpha (float, optional): filter significance level. Defaults to 0.05.
            alpha (float, optional): PCMCI significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.
        """
        super().__init__(data, min_lag, max_lag, verbosity, alpha, resfolder, neglect_only_autodep, clean_cls)
        
        self.f_alpha = f_alpha
        self.sel_method = sel_method
        
        self.validator = myPCMCI(self.alpha, min_lag, max_lag, val_condtest, verbosity, neglect_only_autodep = neglect_only_autodep)       


    def run_filter(self):
        """Run filter method."""
        CP.info("\n")
        CP.info(DASH)
        CP.info("Selecting relevant features among: " + str(self.data.features))
        CP.info("Selection method: " + self.sel_method.name)
        CP.info("Significance level: " + str(self.f_alpha))
        CP.info("Max lag time: " + str(self.max_lag))
        CP.info("Min lag time: " + str(self.min_lag))
        CP.info("Data length: " + str(self.data.T))
       
        self.sel_method.initialise(self.data, self.f_alpha, self.min_lag, self.max_lag, self.CM)
        self.CM = self.sel_method.compute_dependencies()  

   
    def run(self, remove_unneeded = True, nofilter = False) -> DAG:
        """
        Run F-PCMCI.
        
        Args:
            remove_unneeded (bool, optional): Bit to remove unneeded (isolated) variables. Defaults to True.
            nofilter (bool, optional): Bit to run F-PCMCI without filter. Defaults to False.

        Returns:
            DAG: causal model.
        """
        link_assumptions = None
        
        if not nofilter:
            ## 1. FILTER
            self.run_filter()
        
            # list of selected features based on filter dependencies
            self.CM.remove_unneeded_features()
            if not self.CM.features: return None, None
        
            ## 2. VALIDATOR
            # shrink dataframe d by using the filter result
            self.data.shrink(self.CM.features)
        
            # selected links to check by the validator
            link_assumptions = self.CM.get_link_assumptions()
            
            # calculate dependencies on selected links
            f_dag = copy.deepcopy(self.CM)
            
        if self.min_lag != 0:
            self.CM = self.validator.run(self.data, link_assumptions)
        else:
            self.CM = self.validator.run_plus(self.data, link_assumptions)
            
        # list of selected features based on validator dependencies
        if remove_unneeded: self.CM.remove_unneeded_features()
    
        # Saving final causal model
        if not nofilter: self._print_differences(f_dag, self.CM)
        self.save()
        
        if self.resfolder is not None: self.logger.close()
        return self.CM
              
    
    def load(self, res_path):
        """
        Load previously estimated result.

        Args:
            res_path (str): pickle file path.
        """
        with open(res_path, 'rb') as f:
            r = pickle.load(f)
            self.CM = r['causal_model']
            self.f_alpha = r['filter_alpha']
            self.alpha = r['alpha']
            self.dag_path = r['dag_path']
            self.ts_dag_path = r['ts_dag_path']
            
            
    def save(self):
        """Save causal discovery result as pickle file if resfolder is set."""
        if self.respath is not None:
            if self.CM:
                res = dict()
                res['causal_model'] = copy.deepcopy(self.CM)
                res['features'] = copy.deepcopy(self.CM.features)
                res['filter_alpha'] = self.f_alpha
                res['alpha'] = self.alpha
                res['dag_path'] = self.dag_path
                res['ts_dag_path'] = self.ts_dag_path
                with open(self.respath, 'wb') as resfile:
                    pickle.dump(res, resfile)
            else:
                CP.warning("Causal model impossible to save")
    
    
    def _print_differences(self, old_dag : DAG, new_dag : DAG):
        """
        Print difference between old and new dependencies.

        Args:
            old_dep (DAG): old dag.
            new_dep (DAG): new dag.
        """
        # Check difference(s) between validator and filter dependencies
        list_diffs = list()
        tmp = copy.deepcopy(old_dag)
        for t in tmp.g:
            if t not in new_dag.g:
                list_diffs.append(t)
                continue
                
            for s in tmp.g[t].sources:
                if s not in new_dag.g[t].sources:
                    list_diffs.append((s[0], s[1], t))
        
        if list_diffs:
            CP.info("\n")
            CP.info(DASH)
            CP.info("Difference(s):")
            for diff in list_diffs: 
                if type(diff) is tuple:
                    CP.info("Removed (" + str(diff[0]) + " -" + str(diff[1]) +") --> (" + str(diff[2]) + ")")
                else:
                    CP.info(diff + " removed")


"""
This module provides the PAG class.

Classes:
    PAG: class for facilitating the handling and the creation of PAGs.
"""

from pgmpy.models import BayesianNetwork
from itertools import combinations
import copy
from collections import defaultdict

class PAG():
    """PAG class."""
    
    def __init__(self, dag, tau_max, latents) -> None:
        """
        Class constructor.

        Args:
            dag (DAG): DAG to convert.
            tau_max (int): max time lag.
            latents (list[str]): list of latent variables.

        Raises:
            ValueError: latent must be a string
        """
        if not isinstance(latents, list): raise ValueError('latents must be a list')
        self.link_assumptions = dag
        self.tau_max = tau_max
        self.latents = latents
        self.dSepSets = {}

        self.tsDAG = self.createDAG(self.link_assumptions, self.tau_max)
        
        self.pag = self.tsDAG2tsDPAG()
        
    
    def convert2Graph(self) -> dict:
        """
        Convert a PAG to a graph representation.

        Returns:
            dict: Graph representation of a PAG.
        """
        out = {t: {} for t in self.pag}
        for t in self.pag:
            for s in self.pag[t]:
                out[t][(s[0], s[1])] = s[2]
        return out
        
        
    @staticmethod
    def createDAG(link_assumptions, tau_max) -> BayesianNetwork:
        """
        Create a DAG represented by a Baysian Network.

        Args:
            link_assumptions (dict): DAG link assumptions.
            tau_max (int): max time lag.

        Raises:
            ValueError: source not well defined.

        Returns:
            BayesianNetwork: DAG represented by a Baysian Network.
        """
        BN = BayesianNetwork()
        BN.add_nodes_from([(t, -l) for t in link_assumptions.keys() for l in range(0, tau_max)])

        # Edges
        edges = []
        for t in link_assumptions.keys():
            for source in link_assumptions[t]:
                if len(source) == 0: continue
                elif len(source) == 2: s, l = source
                elif len(source) == 3: s, l, _ = source
                else: raise ValueError("Source not well defined")
                edges.append(((s, l), (t, 0)))
                # Add edges across time slices from -1 to -tau_max
                for lag in range(1, tau_max + 1):
                    if l - lag >= -tau_max:
                        edges.append(((s, l - lag), (t, -lag)))
        BN.add_edges_from(edges)
        return BN
    
    
    def alreadyChecked(self, source, target):
        """
        Check if a link has been already checked.

        Args:
            source (str): source node
            target (str): target node

        Returns:
            (bool, tuple): tuple containing if the link has been checked and, if so, their separation set. Otherwise None.
        """
        if (source, target) in self.dSepSets: return True, self.dSepSets[(source, target)]
        elif (target, source) in self.dSepSets: return True, self.dSepSets[(target, source)]
        elif ((source[0], source[1] - target[1]), (target[0], 0)) in self.dSepSets: return True, self.dSepSets[((source[0], source[1] - target[1]), (target[0], 0))]
        elif ((target[0], target[1] - source[1]), (source[0], 0)) in self.dSepSets: return True, self.dSepSets[((target[0], target[1] - source[1]), (source[0], 0))]
        return False, None

    
    def tsDAG2tsDPAG(self) -> dict:
        """
        Convert a DAG to a Time-series DPAG.

        Returns:
            dict: Time-series DPAG.
        """
        self.tsDPAG = {t: [(s[0], s[1], '-->') for s in self.link_assumptions[t] if s[0] not in self.latents] for t in self.link_assumptions.keys() if t not in self.latents}
        
        if len(self.latents) > 0:
            self.ambiguous_links = []
            
            # Separate nodes based on their time index
            time_zero_nodes = []
            other_nodes = []

            for node in self.tsDAG.nodes():
                if node[1] == 0:
                    time_zero_nodes.append(node)
                else:
                    other_nodes.append(node)
            
            for target in time_zero_nodes + other_nodes:
                print(f"Analysing target: {target}")
                if target[0] in self.latents: continue
                tmp = []
                for n in list(self.tsDAG.nodes()):
                    if n[0] != target[0] or n[1] != target[1]:
                        tmp.append(n)
                for p in list(self.tsDAG.predecessors(target)) + list(self.tsDAG.successors(target)):
                    if p in tmp: tmp.remove(p)
                for source in tmp:
                    alreadyChecked, d_sep = self.alreadyChecked(source, target)
                    if alreadyChecked:
                        print(f"\t- {source} ⊥ {target} | {d_sep} ALREADY CHECKED")
                    else:
                        areDsep, d_sep = self.find_d_separators(source, target, self.latents)
                        self.dSepSets[(source, target)] = d_sep
                        print(f"\t- {source} ⊥ {target} | {d_sep}")
                    if areDsep and any(node[0] in self.latents for node in d_sep):
                        if source[0] in self.latents or target[0] in self.latents: continue
                        if target[1] == 0:
                            print(f"\t- SPURIOUS LINK: ({source[0]}, {source[1]}) o-o ({target[0]}, {target[1]})")
                            if (source[0], source[1], 'o-o') not in self.tsDPAG[target[0]]: self.tsDPAG[target[0]].append((source[0], source[1], 'o-o'))
                            if (source, target, 'o-o') not in self.ambiguous_links: self.ambiguous_links.append((source, target, 'o-o'))
                        elif source[1] == 0:
                            print(f"\t- SPURIOUS LINK: ({target[0]}, {target[1]}) o-o ({source[0]}, {source[1]})")
                            if (target[0], target[1], 'o-o') not in self.tsDPAG[source[0]]: self.tsDPAG[source[0]].append((target[0], target[1], 'o-o'))
                            if (target, source, 'o-o') not in self.ambiguous_links: self.ambiguous_links.append((target, source, 'o-o'))
            
            
            print(f"--------------------------------------------------")
            print(f"    Bidirected link due to latent confounders     ")
            print(f"--------------------------------------------------")
            # *(1) Bidirected link between variables confounded by a latent variable  
            # *    if a link between them does not exist already
            confounders = self.find_latent_confounders()
            for confounded in copy.deepcopy(list(confounders.values())):
                for c1 in copy.deepcopy(confounded):
                    tmp = copy.deepcopy(confounded)
                    tmp.remove(c1)
                    for c2 in tmp:
                        if (c1, c2, 'o-o') in self.ambiguous_links:
                            self.update_link_type(c1, c2, '<->')
                            self.ambiguous_links.remove((c1, c2, 'o-o'))
                            print(f"\t- SPURIOUS LINK REMOVED: {c1} o-o {c2}")
                        elif (c2, c1, 'o-o') in self.ambiguous_links:
                            self.update_link_type(c1, c2, '<->')
                            self.ambiguous_links.remove((c2, c1, 'o-o'))
                            print(f"\t- SPURIOUS LINK REMOVED: {c2} o-o {c1}")
                    confounded.remove(c1)
                            
            print(f"--------------------------------------------------")
            print(f"              Collider orientation                ")
            print(f"--------------------------------------------------")
            # *(2) Identify and orient the colliders:
            # *    for any path X – Z – Y where there is no edge between
            # *    X and Y and, Z was never included in the conditioning set ==> X → Z ← Y collider
            colliders = self.find_colliders()
            for ambiguous_link in copy.deepcopy(self.ambiguous_links):
                source, target, linktype = ambiguous_link
                for parent1, collider, parent2 in colliders:
                    if collider == target and (parent1 == source or parent2 == source):
                        if not self.tsDAG.has_edge(parent1, parent2) and not self.tsDAG.has_edge(parent2, parent1):
                            self.update_link_type(parent1, target, '-->')
                            self.update_link_type(parent2, target, '-->')
                            self.ambiguous_links.remove(ambiguous_link)
                            break

            print(f"--------------------------------------------------")
            print(f"Non-collider orientation (orientation propagation)")
            print(f"--------------------------------------------------")
            # *(3) Orient the non-colliders edges (orientation propagation)
            # *    any edge Z – Y part of a partially directed path X → Z – Y,
            # *    where there is no edge between X and Y can be oriented as Z → Y
            for ambiguous_link in copy.deepcopy(self.ambiguous_links):
                triples = self.find_triples_containing_link(ambiguous_link)
                for triple in triples: self.update_link_type(triple[1], triple[2], '-->')
                
        # TODO: (3) Check if cycles are present
        
        else:
            print("No latent variable")
        
        return self.tsDPAG


    def find_colliders(self) -> list:
        """
        Find colliders.

        Returns:
            list: colliders.
        """
        colliders = []
        for node in self.tsDPAG.keys():
            parents = [(p[0], p[1]) for p in self.tsDPAG[node]]
            if len(parents) >= 2:
                for i in range(len(parents)):
                    for j in range(i + 1, len(parents)):
                        parent1 = parents[i]
                        parent2 = parents[j]
                        colliders.append((parent1, (node, 0), parent2))
        return colliders


    def update_link_type(self, parent, target, linktype):
        """
        Update link type.

        Args:
            parent (str): parent node.
            target (str): target node
            linktype (str): link type. E.g. --> or -?>.
        """
        for idx, link in enumerate(self.tsDPAG[target[0]]):
            if link[0] == parent[0] and link[1] == parent[1]:
                self.tsDPAG[target[0]][idx] = (link[0], link[1], linktype)
            
            
    def find_latent_confounders(self) -> dict:
        """
        Find latent confounders.

        Returns:
            dict: latent confounders.
        """
        confounders = {(latent, -t): list(self.tsDAG.successors((latent, -t))) for latent in self.latents for t in range(self.tau_max + 1) if len(list(self.tsDAG.successors((latent, -t)))) > 1}
        
        # Initialize a new dictionary to store unique edges
        shrinked_confounders = defaultdict(list)
        
        # Set to keep track of added edges without considering the time slice
        seen_edges = set()
        
        for key, value in confounders.items():
            # Normalize key by removing the time slice
            key_normalized = key[0]
            
            for v in value:
                # Normalize value by removing the time slice
                v_normalized = v[0]
                
                # Create a tuple of normalized edge
                edge = (key_normalized, v_normalized)
                
                # Check if edge or its reverse has been seen
                if edge not in seen_edges and (v_normalized, key_normalized) not in seen_edges:
                    # If not seen, add to unique edges and mark as seen
                    shrinked_confounders[key].append(v)
                    seen_edges.add(edge)
        
        return shrinked_confounders
    
                
    def find_d_separators(self, source, target):
        """
        Find D-Separation set.

        Args:
            source (str): source node.
            target (str): target node.

        Returns:
            (bool, set): (True, separation set) if source and target are d-separated. Otherwise (False, empty set). 
        """
        paths = self.find_all_paths(source, target)
               
        if len(paths) == 0: 
            return True, set()
        else:
            nodes = set()
            obs_nodes = set()
            for path in paths:
                path.remove(source)
                path.remove(target)
                for node in path:
                    nodes.add(node)
                    if node not in self.latents: obs_nodes.add(node)
                               
            for r in range(len(obs_nodes) + 1):
                for subset in combinations(obs_nodes, r):
                    subset_set = set(subset)
                    if not self.tsDAG.is_dconnected(source, target, subset_set):
                        return True, subset_set
                    
            for r in range(len(nodes) + 1):
                for subset in combinations(nodes, r):
                    subset_set = set(subset)
                    if not self.tsDAG.is_dconnected(source, target, subset_set):
                        return True, subset_set
            
        return False, set()

    def find_triples_containing_link(self, ambiguous_link) -> set:
        """
        Find all triples containing a link.

        Args:
            ambiguous_link (tuple): ambiguous_link

        Returns:
            set: triples containing the specified link.
        """
        pag = self.createDAG(self.tsDPAG, self.tau_max)

        source, target, _ = ambiguous_link
        triples = set()
        
        for n in pag.predecessors(source): 
            if n != target and not pag.has_edge(n, target) and not pag.has_edge(target, n): triples.add((n, source, target))
        for n in pag.predecessors(target): 
            if n != source and not pag.has_edge(n, source) and not pag.has_edge(source, n): triples.add((n, target, source))
        
        return triples
        
    
    # DFS to find all paths
    def find_all_paths(self, start, goal, path=[]) -> list:
        """
        Find all path from start to goal.

        Args:
            start (str): starting node.
            goal (str): goal node.
            path (list, optional): Found paths. Defaults to [].

        Returns:
            list: paths
        """
        path = path + [start]
        if start == goal:
            return [path]
        paths = []
        for node in self.tsDAG.successors(start):
            if node not in path:
                new_paths = self.find_all_paths(node, goal, path)
                for new_path in new_paths:
                    paths.append(new_path)
        for node in self.tsDAG.predecessors(start):
            if node not in path:
                new_paths = self.find_all_paths(node, goal, path)
                for new_path in new_paths:
                    paths.append(new_path)
        return paths

                    




# tau_max = 2

# DAG = {
#     'X_1': [('X_1', -1), ('X_2', -1), ('X_3', 0)],
#     'X_2': [],
#     'X_3': [('X_2', -1), ('X_4', -1)],
#     'X_4': [('X_4', -1)],
# }

# DAG = {
#     'X_1': [('X_2', 0), ('X_1', -1)],
#     'X_2': [],
#     'X_3': [('X_2', -1), ('X_3', -1)],
#     'X_4': [('X_4', -1), ('X_3', -2)],
# }

# p = PAG(DAG, tau_max, ['X_2'])
# print(p.pag)



"""
This module provides the Logger class.

Classes:
    Logger: class responsible for the log.
"""

import sys
from causalflow.basics.utils import cls


class Logger(object):
    """Logger class."""
    
    def __init__(self, path, clean_console = True):
        """
        Class constructor.

        Args:
            path (str): log file path.
            clean_console (bool, optional): clean console flag. Defaults to True.
        """
        self.terminal = sys.stdout
        self.log = open(path, "w")
        if clean_console: cls()

    def write(self, message):
        """
        Write message.

        Args:
            message (str): log msg.
        """
        self.terminal.write(message)
        self.log.write(message)

    def flush(self):
        """python3 compatibility."""
        # this flush method is needed for python 3 compatibility.
        # this handles the flush command by doing nothing.
        # you might want to specify some extra behavior here.
        pass
    
    def close(self):
        """Close logger."""
        sys.stdout = sys.__stdout__
        self.log.close()


"""
This module provides various classes for feature selection analysis.

Classes:
    CTest: support class for handling different feature selection methods.
    SelectionMethod: Abstract class.
"""

from abc import ABC, abstractmethod
from enum import Enum
from contextlib import contextmanager
import sys, os
from causalflow.preprocessing.data import Data
from causalflow.basics.constants import *
from causalflow.CPrinter import CP
from causalflow.graph.DAG import DAG


class CTest(Enum):
    """CTest Enumerator."""
    
    Corr = "Correlation"
    MI = "Mutual Information"
    TE = "Transfer Entropy"


@contextmanager
def _suppress_stdout():
    """Suppress stdout."""
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:  
            yield
        finally:
            sys.stdout = old_stdout


class SelectionMethod(ABC):
    """SelectionMethod abstract class."""
    
    def __init__(self, ctest):
        """
        Class constructor.

        Args:
            ctest (CTest): Feature Selection method's name.
        """
        self.ctest = ctest
        self.data = None
        self.alpha = None
        self.min_lag = None
        self.max_lag = None
        self.result = None


    @property
    def name(self):
        """
        Return Selection Method name.

        Returns:
            (str): Selection Method name.
        """
        return self.ctest.value


    def initialise(self, data: Data, alpha, min_lag, max_lag, graph):
        """
        Initialise the selection method.

        Args:
            data (Data): Data.
            alpha (float): significance threshold.
            min_lag (int): min lag time.
            max_lag (int): max lag time.
            graph (DAG): initial DAG (empty).
        """
        self.data = data
        self.alpha = alpha
        self.min_lag = min_lag
        self.max_lag = max_lag
        self.result = graph


    @abstractmethod
    def compute_dependencies(self) -> DAG:
        """Abstract method."""
        pass
    

    def _prepare_ts(self, target, lag, apply_lag = True, consider_autodep = True):
        """
        Prepare the dataframe to the analysis.

        Args:
            target (str): name target var
            lag (int): lag time to apply
            apply_lag (bool, optional): True if you want to apply the lag, False otherwise. Defaults to True.
            consider_autodep (bool, optional): True if you want to consider autodependecy check. Defaults to True.

        Returns:
            tuple(DataFrame, DataFrame): source and target dataframe.
        """
        if not consider_autodep:
            if apply_lag:
                Y = self.data.d[target][lag:]
                X = self.data.d.loc[:, self.data.d.columns != target][:-lag]
            else:
                Y = self.data.d[target]
                X = self.data.d.loc[:, self.data.d.columns != target]
        else:
            if apply_lag:
                Y = self.data.d[target][lag:]
                X = self.data.d[:-lag]
            else:
                Y = self.data.d[target]
                X = self.data.d
        return X, Y


    def _add_dependency(self, t, s, score, pval, lag):
        """
        Add dependency from source (s) to target (t) specifying the score, pval and the lag.

        Args:
            t (str): target feature name.
            s (str): source feature name.
            score (float): selection method score.
            pval (float): pval associated to the dependency.
            lag (int): lag time of the dependency.
        """
        self.result.add_source(t, s, score, pval, lag)
        
        str_s = "(" + s + " -" + str(lag) + ")"
        str_t = "(" + t + ")"

        CP.info("\tlink: " + str_s + " -?> " + str_t)
        CP.info("\t|val = " + str(round(score,3)) + " |pval = " + str(str(round(pval,3))))
        # CP.info("\n")





"""
This module provides the Data class.

Classes:
    Data: public class for handling data used for the causal discovery.
"""

import numpy as np
import pandas as pd
from matplotlib import gridspec
import matplotlib.pyplot as plt
from sklearn.preprocessing import StandardScaler
from causalflow.preprocessing.Subsampler import Subsampler
from causalflow.preprocessing.subsampling_methods.SubsamplingMethod import SubsamplingMethod


class Data():
    """Data class manages the preprocess of the data before the causal analysis."""
    
    def __init__(self, data, vars = None, fill_nan = True, stand = False, subsampling : SubsamplingMethod = None, show_subsampling = False):
        """
        Class constructor.

        Args:
            data (str/DataFrame/np.array): it can be a string specifing the path of a csv file to load/pandas.DataFrame/numpy.array.
            vars (list(str), optional): List containing variable names. If unset then, 
                if data = (str/DataFrame) vars = data columns name elif data = np.array vars = [X_0 .. X_N]
                Defaults to None.
            fill_nan (bool, optional): Fill NaNs bit. Defaults to True.
            stand (bool, optional): Standardization bit. Defaults to False.
            subsampling (SubsamplingMethod, optional): Subsampling method. If None not active. Defaults to None.
            show_subsampling (bool, optional): If True shows subsampling result. Defaults to False.

        Raises:
            TypeError: if data is not str - DataFrame - ndarray.
        """
        # Data handling
        if type(data) == np.ndarray:
            self.d = pd.DataFrame(data)
            if vars is None: self.d.columns = list(['X_' + str(f) for f in range(len(self.d.columns))])
        elif type(data) == pd.DataFrame:
            self.d = data
        elif type(data) == str:
            self.d = pd.read_csv(data)
        else:
            raise TypeError("data field not in the correct type\ndata must be one of the following type:\n- numpy.ndarray\n- pandas.DataFrame\n- .csv path")
            
        
        # Columns name handling
        if vars is not None:
            self.d.columns = list(vars)
                
        
        self.orig_features = self.features
        self.orig_pretty_features = self.pretty_features
        self.orig_N = self.N
        self.orig_T = len(self.d)

        # Filling NaNs
        if fill_nan:
            if self.d.isnull().values.any():
                self.d.fillna(inplace=True, method="ffill")
                self.d.fillna(inplace=True, method="bfill")

        # Subsampling data
        if subsampling is not None:
            subsampler = Subsampler(self.d, ss_method = subsampling)
            self.d = pd.DataFrame(subsampler.subsample(), columns = self.features)
            if show_subsampling: subsampler.plot_subsampled_data()

        # Standardize data
        if stand:
            scaler = StandardScaler()
            scaler = scaler.fit(self.d)
            self.d = pd.DataFrame(scaler.transform(self.d), columns = self.features)
        
    @property  
    def features(self):
        """
        Return list of features.

        Returns:
            list(str): list of feature names.
        """
        return list(self.d.columns)

    @property
    def pretty_features(self):
        """
        Return list of features with LATEX symbols.
                
        Returns:
            list(str): list of feature names.
        """
        return [r'$' + str(v) + '$' for v in self.d.columns]
    
    @property
    def N(self):
        """
        Number of features.
        
        Returns:
            (int): number of features.
        """
        return len(self.d.columns)

    @property
    def T(self):
        """
        Dataframe length.
        
        Returns:
            (int): dataframe length.
        """
        return len(self.d)
                       
            
    def shrink(self, selected_features):
        """
        Shrink dataframe d on the selected features.

        Args:
            selected_features (list(str)): list of variables.
        """
        self.d = self.d[selected_features]
        
                    
    def plot_timeseries(self, savefig = None):
        """
        Plot timeseries data.
        
        Args:
            savefig (str): figure path.
        """
        # Create grid
        gs = gridspec.GridSpec(self.N, 1)

        # Time vector
        T = list(range(self.T))

        plt.figure()
        for i in range(0, self.d.shape[1]):
            ax = plt.subplot(gs[i, 0])
            plt.plot(T, self.d.values[:, i], color = 'tab:red')
            plt.ylabel(str(self.pretty_features[i]))

        if savefig is not None:
            plt.savefig(savefig)
        else:
            plt.show()
            
            
    def save_csv(self, csvpath):
        """
        Save timeseries data into a CSV file.
        
        Args:
            csvpath (str): CSV path.
        """
        self.d.to_csv(csvpath, index=False)


"""
This module provides the DAG class.

Classes:
    DAG: class for facilitating the handling and the creation of DAGs.
"""
    
import copy
import numpy as np
from causalflow.graph.Node import Node
from causalflow.basics.constants import *
from matplotlib import pyplot as plt
import networkx as nx
from causalflow.graph.netgraph import Graph
import re

class DAG():
    """DAG class."""
    
    def __init__(self, var_names, min_lag, max_lag, neglect_autodep = False, scm = None):
        """
        DAG constructor.

        Args:
            var_names (list): variable list.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            neglect_autodep (bool, optional): bit to neglect nodes when they are only autodependent. Defaults to False.
            scm (dict, optional): Build the DAG for SCM. Defaults to None.
        """
        self.g = {var: Node(var, neglect_autodep) for var in var_names}
        self.neglect_autodep = neglect_autodep
        self.sys_context = dict()
        self.min_lag = min_lag
        self.max_lag = max_lag
        
        if scm is not None:
            for t in scm:
                    for s in scm[t]: 
                        if len(s) == 2:
                            self.add_source(t, s[0], 0.3, 0, s[1])
                        elif len(s) == 3:
                            self.add_source(t, s[0], 0.3, 0, s[1], s[2])


    @property
    def features(self) -> list:
        """
        Return features list.

        Returns:
            list: Features list.
        """
        return list(self.g.keys())
    
    
    @property
    def pretty_features(self) -> list:
        """
        Return list of features with LaTeX symbols.
                
        Returns:
            list(str): list of feature names.
        """
        return [r'$' + str(v) + '$' for v in self.g.keys()]

    
    @property
    def autodep_nodes(self) -> list:
        """
        Return the autodependent nodes list.

        Returns:
            list: Autodependent nodes list.
        """
        autodeps = list()
        for t in self.g:
            # NOTE: I commented this because I want to check all the auto-dep nodes with obs data
            # if self.g[t].is_autodependent and self.g[t].intervention_node: autodeps.append(t)
            if self.g[t].is_autodependent: autodeps.append(t)
        return autodeps
    
    
    @property
    def interventions_links(self) -> list:
        """
        Return the intervention links list.

        Returns:
            list: Intervention link list.
        """
        int_links = list()
        for t in self.g:
            for s in self.g[t].sources:
                if self.g[s[0]].intervention_node:
                    int_links.append((s[0], s[1], t))
        return int_links
    
    
    @property
    def max_auto_score(self) -> float:
        """
        Return maximum score of an auto-dependency link.

        Returns:
            float: maximum score of an auto-dependency link.
        """
        return max([self.g[t].sources[self.g[t].get_max_autodependent][SCORE] for t in self.g if self.g[t].is_autodependent])
    
    
    @property
    def max_cross_score(self) -> float:
        """
        Return maximum score of an cross-dependency link.

        Returns:
            float: maximum score of an cross-dependency link.
        """
        return max([self.g[t].sources[s][SCORE] if self.g[t].sources[s][SCORE] != float('inf') else 1 for t in self.g for s in self.g[t].sources if t != s[0]])
      
    
    def add_source(self, t, s, score, pval, lag, mode = LinkType.Directed.value):
        """
        Add source node to a target node.

        Args:
            t (str): target node name.
            s (str): source node name.
            score (float): dependency score.
            pval (float): dependency p-value.
            lag (int): dependency lag.
            mode (LinkType): link type. E.g., Directed -->
        """
        self.g[t].sources[(s, abs(lag))] = {SCORE: score, PVAL: pval, TYPE: mode}
        self.g[s].children.append(t)
       
        
    def del_source(self, t, s, lag):
        """
        Remove source node from a target node.

        Args:
            t (str): target node name.
            s (str): source node name.
            lag (int): dependency lag.
        """
        del self.g[t].sources[(s, lag)]
        self.g[s].children.remove(t)
        
        
    def remove_unneeded_features(self):
        """Remove isolated nodes."""
        tmp = copy.deepcopy(self.g)
        for t in self.g.keys():
            if self.g[t].is_isolated: 
                if self.g[t].intervention_node: del tmp[self.g[t].associated_context]
                del tmp[t]
        self.g = tmp
                          
    
    def add_context(self):
        """Add context variables."""
        for sys_var, context_var in self.sys_context.items():
            if sys_var in self.features:
                
                # Adding context var to the graph
                self.g[context_var] = Node(context_var, self.neglect_autodep)
                
                # Adding context var to sys var
                self.g[sys_var].intervention_node = True
                self.g[sys_var].associated_context = context_var
                self.add_source(sys_var, context_var, 1, 0, 0)
                
        # NOTE: bi-directed link contemporanous link between context vars
        for sys_var, context_var in self.sys_context.items():
            if sys_var in self.features:
                other_context = [value for value in self.sys_context.values() if value != context_var and value in self.features]
                for other in other_context: self.add_source(context_var, other, 1, 0, 0)
        
                    
    def remove_context(self):
        """Remove context variables."""
        for sys_var, context_var in self.sys_context.items():
            if sys_var in self.g:
                # Removing context var from sys var
                # self.g[sys_var].intervention_node = False
                self.g[sys_var].associated_context = None
                self.del_source(sys_var, context_var, 0)
                    
                # Removing context var from dag
                del self.g[context_var]
                                                                  
                
    def get_link_assumptions(self, autodep_ok = False) -> dict:
        """
        Return link assumption dictionary.

        Args:
            autodep_ok (bool, optional): If true, autodependecy link assumption = -->. Otherwise -?>. Defaults to False.

        Returns:
            dict: link assumption dictionary.
        """
        link_assump = {self.features.index(f): dict() for f in self.features}
        for t in self.g:
            for s in self.g[t].sources:
                if autodep_ok and s[0] == t: # NOTE: new condition added in order to not control twice the autodependency links
                    link_assump[self.features.index(t)][(self.features.index(s[0]), -abs(s[1]))] = '-->'
                    
                elif s[0] not in list(self.sys_context.values()):
                    if s[1] == 0 and (t, 0) in self.g[s[0]].sources:
                        link_assump[self.features.index(t)][(self.features.index(s[0]), 0)] = 'o-o'
                    elif s[1] == 0 and (t, 0) not in self.g[s[0]].sources:
                        link_assump[self.features.index(t)][(self.features.index(s[0]),0)] = '-?>'
                        link_assump[self.features.index(s[0])][(self.features.index(t), 0)] = '<?-'
                    elif s[1] > 0:
                        link_assump[self.features.index(t)][(self.features.index(s[0]), -abs(s[1]))] = '-?>'
                    
                elif t in self.sys_context.keys() and s[0] == self.sys_context[t]:
                    link_assump[self.features.index(t)][(self.features.index(s[0]), -abs(s[1]))] = '-->'
                    
        return link_assump
   
    
    def make_pretty(self) -> dict:
        """
        Make variables' names pretty, i.e. $ varname $ with '{' after '_' and '}' at the end of the string.

        Returns:
            dict: pretty DAG.
        """
        def prettify(name):
            return '$' + re.sub(r'_(\w+)', r'_{\1}', name) + '$'
        
        pretty = dict()
        for t in self.g:
            p_t = prettify(t)
            pretty[p_t] = copy.deepcopy(self.g[t])
            pretty[p_t].name = p_t
            pretty[p_t].children = [prettify(c) for c in self.g[t].children]
            for s in self.g[t].sources:
                del pretty[p_t].sources[s]
                p_s = prettify(s[0])
                pretty[p_t].sources[(p_s, s[1])] = {
                    SCORE: self.g[t].sources[s][SCORE],
                    PVAL: self.g[t].sources[s][PVAL],
                    TYPE: self.g[t].sources[s][TYPE]
                }
        return pretty
    
    
    def __add_edge(self, min_width, max_width, min_score, max_score, edges, edge_width, arrows, r, t, s, s_node, t_node):
        """
        Add edge to a graph. Support method for dag and ts_dag.

        Args:
            min_width (int): minimum linewidth. Defaults to 1.
            max_width (int): maximum linewidth. Defaults to 5.
            min_score (int): minimum score range. Defaults to 0.
            max_score (int): maximum score range. Defaults to 1.
            edges (list): list of edges.
            edge_width (dict): dictionary containing the width for each edge of the graph.
            arrows (dict): dictionary containing a bool for each edge of the graph describing if the edge is directed or not.
            r (DAG): DAG.
            t (str or tuple): target node.
            s (str or tuple): source node.
            s_node (str): source node.
            t_node (str): target node.

        Raises:
            ValueError: link type associated to this edge not included in our LinkType list.
        """
        edges.append((s_node, t_node))
        score = r.g[t].sources[s][SCORE] if r.g[t].sources[s][SCORE] != float('inf') else 1
        edge_width[(s_node, t_node)] = self.__scale(score, min_width, max_width, min_score, max_score)
        
        if r.g[t].sources[s][TYPE] == LinkType.Directed.value:
            arrows[(s_node, t_node)] = {'h':'>', 't':''}
            
        elif r.g[t].sources[s][TYPE] == LinkType.Bidirected.value:
            edges.append((t_node, s_node))
            edge_width[(t_node, s_node)] = self.__scale(score, min_width, max_width, min_score, max_score)
            arrows[(t_node, s_node)] = {'h':'>', 't':''}
            arrows[(s_node, t_node)] = {'h':'>', 't':''}
            
        elif r.g[t].sources[s][TYPE] == LinkType.HalfUncertain.value:
            arrows[(s_node, t_node)] = {'h':'>', 't':'o'}
            
        elif r.g[t].sources[s][TYPE] == LinkType.Uncertain.value:
            arrows[(s_node, t_node)] = {'h':'o', 't':'o'}
        
        else:
            raise ValueError(f"{r.g[t].sources[s][TYPE]} not included in LinkType")
             
    
    def dag(self,
        node_layout='dot',
        min_auto_width=0.25, 
        max_auto_width=0.75,
        min_cross_width=1, 
        max_cross_width=5,
        node_size=8, 
        node_color='orange',
        edge_color='grey',
        tail_color='black',
        font_size=8,
        label_type=LabelType.Lag,
        save_name=None,
        img_extention=ImageExt.PNG):
        """
        Build a dag, first with contemporaneous links, then lagged links.

        Args:
            node_layout (str, optional): Node layout. Defaults to 'dot'.
            min_auto_width (float, optional): minimum border linewidth. Defaults to 0.25.
            max_auto_width (float, optional): maximum border linewidth. Defaults to 0.75.
            min_cross_width (float, optional): minimum edge linewidth. Defaults to 1.
            max_cross_width (float, optional): maximum edge linewidth. Defaults to 5.
            node_size (int, optional): node size. Defaults to 8.
            node_color (str, optional): node color. Defaults to 'orange'.
            edge_color (str, optional): edge color for contemporaneous links. Defaults to 'grey'.
            tail_color (str, optional): tail color. Defaults to 'black'.
            font_size (int, optional): font size. Defaults to 8.
            label_type (LabelType, optional): Show the lag time (LabelType.Lag), the strength (LabelType.Score), or no labels (LabelType.NoLabels). Default LabelType.Lag.
            save_name (str, optional): Filename path. If None, plot is shown and not saved. Defaults to None.
            img_extention (ImageExt, optional): Image Extension. Defaults to PNG.
        """
        r = copy.deepcopy(self)
        r.g = r.make_pretty()

        Gcont = nx.DiGraph()
        Glag = nx.DiGraph()

        # 1. Nodes definition
        Gcont.add_nodes_from(r.g.keys())
        Glag.add_nodes_from(r.g.keys())
        
        # 2. Nodes border definition
        border = dict()
        for t in r.g:
            border[t] = 0
            if r.g[t].is_autodependent:
                border[t] = max(self.__scale(r.g[t].sources[r.g[t].get_max_autodependent][SCORE], 
                                             min_auto_width, max_auto_width, 
                                             0, self.max_auto_score), 
                                border[t])
        
        # 3. Nodes border label definition
        node_label = None
        if label_type == LabelType.Lag or label_type == LabelType.Score:
            node_label = {t: [] for t in r.g.keys()}
            for t in r.g:
                if r.g[t].is_autodependent:
                    autodep = r.g[t].get_max_autodependent
                    if label_type == LabelType.Lag:
                        node_label[t].append(autodep[1])
                    elif label_type == LabelType.Score:
                        node_label[t].append(round(r.g[t].sources[autodep][SCORE], 3))
                node_label[t] = ",".join(str(s) for s in node_label[t])

        # 3. Edges definition
        cont_edges = []
        cont_edge_width = dict()
        cont_arrows = {}
        lagged_edges = []
        lagged_edge_width = dict()
        lagged_arrows = {}
        
        for t in r.g:
            for s in r.g[t].sources:
                if t != s[0]:  # skip self-loops
                    if s[1] == 0:  # Contemporaneous link (no lag)
                        self.__add_edge(min_cross_width, max_cross_width, 0, self.max_cross_score,
                                        cont_edges, cont_edge_width, cont_arrows, r, t, s, s[0], t)
                    else:  # Lagged link
                        self.__add_edge(min_cross_width, max_cross_width, 0, self.max_cross_score,
                                        lagged_edges, lagged_edge_width, lagged_arrows, r, t, s, s[0], t)
                        
        Gcont.add_edges_from(cont_edges)
        Glag.add_edges_from(lagged_edges)


        fig, ax = plt.subplots(figsize=(8, 6))

        # 4. Edges label definition
        edge_label = None
        lagged_edge_label = None
        if label_type == LabelType.Lag or label_type == LabelType.Score:
            edge_label = {(s[0], t): [] for t in r.g for s in r.g[t].sources if t != s[0] and s[1] == 0}
            lagged_edge_label = {(s[0], t): [] for t in r.g for s in r.g[t].sources if t != s[0] and s[1] != 0}
            for t in r.g:
                for s in r.g[t].sources:
                    if t != s[0]:
                        if s[1] == 0:  # Contemporaneous
                            if label_type == LabelType.Lag:
                                edge_label[(s[0], t)].append(s[1])
                            elif label_type == LabelType.Score:
                                edge_label[(s[0], t)].append(round(r.g[t].sources[s][SCORE], 3))
                        else:  # Lagged
                            if label_type == LabelType.Lag:
                                lagged_edge_label[(s[0], t)].append(s[1])
                            elif label_type == LabelType.Score:
                                lagged_edge_label[(s[0], t)].append(round(r.g[t].sources[s][SCORE], 3))
            for k in edge_label.keys():
                edge_label[k] = ",".join(str(s) for s in edge_label[k])
            for k in lagged_edge_label.keys():
                lagged_edge_label[k] = ",".join(str(s) for s in lagged_edge_label[k])

        # 5. Draw graph - contemporaneous
        if cont_edges:
            a = Graph(Gcont,
                    node_layout=node_layout,
                    node_size=node_size,
                    node_color=node_color,
                    node_labels=None,
                    node_edge_width=border,
                    node_label_fontdict=dict(size=font_size),
                    node_edge_color=edge_color,
                    node_label_offset=0.05,
                    node_alpha=1,

                    arrows=cont_arrows,
                    edge_layout='straight',
                    edge_label=label_type != LabelType.NoLabels,
                    edge_labels=edge_label,
                    edge_label_fontdict=dict(size=font_size),
                    edge_color=edge_color,
                    tail_color=tail_color,
                    edge_width=cont_edge_width,
                    edge_alpha=1,
                    edge_zorder=1,
                    edge_label_position=0.35)

            nx.draw_networkx_labels(Gcont,
                                    pos=a.node_positions,
                                    labels={n: n for n in Glag},
                                    font_size=font_size)

        # 6. Draw graph - lagged
        if lagged_edges:
            a = Graph(Glag,
                    node_layout=a.node_positions,
                    node_size=node_size,
                    node_color=node_color,
                    node_labels=node_label,
                    node_edge_width=border,
                    node_label_fontdict=dict(size=font_size),
                    node_edge_color=edge_color,
                    node_label_offset=0.05,
                    node_alpha=1,

                    arrows=cont_arrows,
                    edge_layout='curved',
                    edge_label=label_type != LabelType.NoLabels,
                    edge_labels=lagged_edge_label,
                    edge_label_fontdict=dict(size=font_size),
                    edge_color=edge_color,
                    tail_color=tail_color,
                    edge_width=lagged_edge_width,
                    edge_alpha=1,
                    edge_zorder=1,
                    edge_label_position=0.35)

        # 7. Plot or save
        if save_name is not None:
            plt.savefig(save_name + img_extention.value, dpi=300)
        else:
            plt.show()
          
   
    def ts_dag(self,
               min_cross_width = 1, 
               max_cross_width = 5,
               node_size = 8,
               x_disp = 1.5,
               y_disp = 0.2,
               node_color = 'orange',
               edge_color = 'grey',
               tail_color = 'black',
               font_size = 8,
               save_name = None,
               img_extention = ImageExt.PNG):
        """
        Build a timeseries dag.

        Args:
            min_cross_width (int, optional): minimum linewidth. Defaults to 1.
            max_cross_width (int, optional): maximum linewidth. Defaults to 5.
            node_size (int, optional): node size. Defaults to 8.
            x_disp (float, optional): node displacement along x. Defaults to 1.5.
            y_disp (float, optional): node displacement along y. Defaults to 0.2.
            node_color (str/list, optional): node color. 
                                             If a string, all the nodes will have the same colour. 
                                             If a list (same dimension of features), each colour will have the specified colour.
                                             Defaults to 'orange'.
            edge_color (str, optional): edge color. Defaults to 'grey'.
            tail_color (str, optional): tail color. Defaults to 'black'.
            font_size (int, optional): font size. Defaults to 8.
            save_name (str, optional): Filename path. If None, plot is shown and not saved. Defaults to None.
            img_extention (ImageExt, optional): Image Extension. Defaults to PNG.
        """
        r = copy.deepcopy(self)
        r.g = r.make_pretty()

        Gcont = nx.DiGraph()
        Glagcross = nx.DiGraph()
        Glagauto = nx.DiGraph()

        # 1. Nodes definition
        if isinstance(node_color, list):
            node_c = dict()
        else:
            node_c = node_color
        for i in range(len(self.features)):
            for j in range(self.max_lag + 1):
                Glagauto.add_node((j, i))
                Glagcross.add_node((j, i))
                Gcont.add_node((j, i))
                if isinstance(node_color, list): node_c[(j, i)] = node_color[abs(i - (len(r.g.keys()) - 1))]
                
        pos = {n : (n[0]*x_disp, n[1]*y_disp) for n in Glagauto.nodes()}
        scale = max(pos.values())

        # 2. Edges definition
        cont_edges = list()
        cont_edge_width = dict()
        cont_arrows = dict()
        lagged_cross_edges = list()
        lagged_cross_edge_width = dict()
        lagged_cross_arrows = dict()
        lagged_auto_edges = list()
        lagged_auto_edge_width = dict()
        lagged_auto_arrows = dict()

        for t in r.g:
            for s in r.g[t].sources:
                s_index = len(r.g.keys())-1 - list(r.g.keys()).index(s[0])
                t_index = len(r.g.keys())-1 - list(r.g.keys()).index(t)
                
                # 2.1. Contemporaneous edges definition
                if s[1] == 0:
                    for i in range(self.max_lag + 1):
                        s_node = (i, s_index)
                        t_node = (i, t_index)
                        self.__add_edge(min_cross_width, max_cross_width, 0, self.max_cross_score, 
                                        cont_edges, cont_edge_width, cont_arrows, r, t, s, 
                                        s_node, t_node)
                        
                else:
                    s_lag = self.max_lag - s[1]
                    t_lag = self.max_lag
                    while s_lag >= 0:
                        s_node = (s_lag, s_index)
                        t_node = (t_lag, t_index)
                        # 2.2. Lagged cross edges definition
                        if s[0] != t:
                            self.__add_edge(min_cross_width, max_cross_width, 0, self.max_cross_score, 
                                            lagged_cross_edges, lagged_cross_edge_width, lagged_cross_arrows, r, t, s, 
                                            s_node, t_node)
                        # 2.3. Lagged auto edges definition
                        else:
                            self.__add_edge(min_cross_width, max_cross_width, 0, self.max_cross_score, 
                                            lagged_auto_edges, lagged_auto_edge_width, lagged_auto_arrows, r, t, s, 
                                            s_node, t_node)
                        s_lag -= 1
                        t_lag -= 1
                    
        Gcont.add_edges_from(cont_edges)
        Glagcross.add_edges_from(lagged_cross_edges)
        Glagauto.add_edges_from(lagged_auto_edges)

        fig, ax = plt.subplots(figsize=(8,6))
        edge_layout = self.__get_fixed_edges(ax, x_disp, Gcont, node_size, pos, node_c, font_size, 
                                             cont_arrows, edge_color, tail_color, cont_edge_width, scale)
        
        # 3. Label definition
        for n in Gcont.nodes():
            if n[0] == 0:
                ax.text(pos[n][0]-0.1, pos[n][1], list(r.g.keys())[len(r.g.keys()) - 1 - n[1]], horizontalalignment='center', verticalalignment='center', fontsize=font_size)

        # 4. Time line text drawing
        pos_tau = set([pos[p][0] for p in pos])
        max_y = max([pos[p][1] for p in pos])
        for p in pos_tau:
            if abs(int(p/x_disp) - self.max_lag) == 0:
                ax.text(p, max_y + 0.1, r"$t$", horizontalalignment='center', fontsize=font_size)
            else:
                ax.text(p, max_y + 0.1, r"$t-" + str(abs(int(p/x_disp) - self.max_lag)) + "$", horizontalalignment='center', fontsize=font_size)
        
        # 5. Draw graph - contemporaneous
        if cont_edges:
            a = Graph(Gcont,
                    node_layout={p : np.array(pos[p]) for p in pos},
                    node_size=node_size,
                    node_color=node_c,
                    node_edge_width=0,
                    node_label_fontdict=dict(size=font_size),
                    node_label_offset=0,
                    node_alpha=1,

                    arrows=cont_arrows,
                    edge_layout=edge_layout,
                    edge_label=False,
                    edge_color=edge_color,
                    tail_color=tail_color,
                    edge_width=cont_edge_width,
                    edge_alpha=1,
                    edge_zorder=1,
                    scale = (scale[0] + 2, scale[1] + 2))

        # 6. Draw graph - lagged cross
        if lagged_cross_edges:
            a = Graph(Glagcross,
                    node_layout={p : np.array(pos[p]) for p in pos},
                    node_size=node_size,
                    node_color=node_c,
                    node_edge_width=0,
                    node_label_fontdict=dict(size=font_size),
                    node_label_offset=0,
                    node_alpha=1,

                    arrows=lagged_cross_arrows,
                    edge_layout='curved',
                    edge_label=False,
                    edge_color=edge_color,
                    tail_color=tail_color,
                    edge_width=lagged_cross_edge_width,
                    edge_alpha=1,
                    edge_zorder=1,
                    scale = (scale[0] + 2, scale[1] + 2))
            
        # 7. Draw graph - lagged auto
        if lagged_auto_edges:
            a = Graph(Glagauto,
                    node_layout={p : np.array(pos[p]) for p in pos},
                    node_size=node_size,
                    node_color=node_c,
                    node_edge_width=0,
                    node_label_fontdict=dict(size=font_size),
                    node_label_offset=0,
                    node_alpha=1,

                    arrows=lagged_auto_arrows,
                    edge_layout='straight',
                    edge_label=False,
                    edge_color=edge_color,
                    tail_color=tail_color,
                    edge_width=lagged_auto_edge_width,
                    edge_alpha=1,
                    edge_zorder=1,
                    scale = (scale[0] + 2, scale[1] + 2))
        
        # 7. Plot or save
        if save_name is not None:
            plt.savefig(save_name + img_extention.value, dpi = 300)
        else:
            plt.show()
            
            
    def __get_fixed_edges(self, ax, x_disp, Gcont, node_size, pos, node_c, font_size, cont_arrows, edge_color, tail_color, cont_edge_width, scale) -> dict:
        """
        Fix edge paths at t-tau_max.

        Args:
            ax (Axes): figure axis.
            x_disp (float): node displacement along x. Defaults to 1.5.
            Gcont (DiGraph): Direct Graph containing only contemporaneous links.
            node_size (int): node size.
            pos (dict): node layout.
            node_c (str/list, optional): node color. 
                                         If a string, all the nodes will have the same colour. 
                                         If a list (same dimension of features), each colour will have the specified colour.
            font_size (int): font size.
            cont_arrows (dict): edge-arrows dictionary .
            edge_color (str): edge color.
            tail_color (str): tail color.
            cont_edge_width (dict): edge-width dictionary.
            scale (tuple): graph scale.

        Returns:
            dict: new edge paths
        """
        a = Graph(Gcont,
                  node_layout={p : np.array(pos[p]) for p in pos},
                  node_size=node_size,
                  node_color=node_c,
                  node_edge_width=0,
                  node_label_fontdict=dict(size=font_size),
                  node_label_offset=0,
                  node_alpha=1,

                  arrows=cont_arrows,
                  edge_layout='curved',
                  edge_label=False,
                  edge_color=edge_color,
                  tail_color=tail_color,
                  edge_width=cont_edge_width,
                  edge_alpha=1,
                  edge_zorder=1,
                  scale = (scale[0] + 2, scale[1] + 2))
        res = copy.deepcopy(a.edge_layout.edge_paths)
        for edge, edge_path in a.edge_layout.edge_paths.items():
            if edge[0][0] == self.max_lag and edge[1][0] == self.max_lag: # t
                for t in range(0, self.max_lag):
                    for fixed_edge in a.edge_layout.edge_paths.keys():
                        if fixed_edge == edge: continue
                        if fixed_edge[0][0] == t and fixed_edge[0][1] == edge[0][1] and fixed_edge[1][0] == t and fixed_edge[1][1] == edge[1][1]:
                            res[fixed_edge] = edge_path - np.array([(self.max_lag - t)*x_disp,0])*np.ones_like(a.edge_layout.edge_paths[edge])
            # if edge[0][0] == 0 and edge[1][0] == 0: # t-tau_max
            #     for shifted_edge, shifted_edge_path in a.edge_layout.edge_paths.items():
            #         if shifted_edge == edge: continue
            #         if shifted_edge[0][0] == self.max_lag and shifted_edge[0][1] == edge[0][1] and shifted_edge[1][0] == self.max_lag and shifted_edge[1][1] == edge[1][1]:
            #             res[edge] = shifted_edge_path - np.array([x_disp,0])*np.ones_like(a.edge_layout.edge_paths[shifted_edge])
        ax.clear()              
        return res
        
    def __scale(self, score, min_width, max_width, min_score = 0, max_score = 1):
        """
        Scale the score of the cause-effect relationship strength to a linewitdth.

        Args:
            score (float): score to scale.
            min_width (float): minimum linewidth.
            max_width (float): maximum linewidth.
            min_score (int, optional): minimum score range. Defaults to 0.
            max_score (int, optional): maximum score range. Defaults to 1.

        Returns:
            (float): scaled score.
        """
        return ((score - min_score) / (max_score - min_score)) * (max_width - min_width) + min_width


    def get_skeleton(self) -> np.array:
        """
        Return skeleton matrix.
        
        Skeleton matrix is composed by 0 and 1.
        1 <- if there is a link from source to target 
        0 <- if there is not a link from source to target 

        Returns:
            np.array: skeleton matrix
        """
        r = np.full((len(self.features), len(self.features), self.max_lag + 1), '', dtype=object)
        for t in self.g.keys():
            for s in self.g[t].sources:
                r[self.features.index(t), self.features.index(s[0])][s[1]] = 1
        return np.array(r)
    
    
    def get_val_matrix(self) -> np.array:
        """
        Return val matrix.
        
        Val matrix contains information about the strength of the links componing the causal model.

        Returns:
            np.array: val matrix.
        """
        r = np.zeros((len(self.features), len(self.features), self.max_lag + 1))
        for t in self.g.keys():
            for s, info in self.g[t].sources.items():
                    r[self.features.index(t), self.features.index(s[0])][s[1]] = info[SCORE]
        return np.array(r)


    def get_pval_matrix(self) -> np.array:
        """
        Return pval matrix.
        
        Pval matrix contains information about the pval of the links componing the causal model.
        
        Returns:
            np.array: pval matrix
        """
        r = np.zeros((len(self.features), len(self.features), self.max_lag + 1))
        for t in self.g.keys():
            for s, info in self.g[t].sources.items():
                r[self.features.index(t), self.features.index(s[0])][s[1]] = info[PVAL]
        return np.array(r)
    
    
    def get_graph_matrix(self) -> np.array:
        """
        Return graph matrix.
        
        Graph matrix contains information about the link type. E.g., -->, <->, ..
        
        Returns:
            np.array: graph matrix.
        """
        r = np.full((len(self.features), len(self.features), self.max_lag + 1), '', dtype=object)
        for t in self.g.keys():
            for s, info in self.g[t].sources.items():
                r[self.features.index(t), self.features.index(s[0])][s[1]] = info[TYPE]
        return np.array(r)
    
    
    def get_Adj(self, indexed = False) -> dict:   
        """
        Return Adjacency dictionary.
        
        If indexed = True: example {0: [(0, -1), (1, -2)], 1: [], ...}
        If indexed = False: example {"X_0": [(X_0, -1), (X_1, -2)], "X_1": [], ...}
        
        Args:
            indexed (bool, optional): If true, returns the SCM with index instead of variables' names. Otherwise it uses variables' names. Defaults to False.
        
        Returns:
            dict: SCM.
        """
        if not indexed:
            scm = {v: list() for v in self.features}
            for t in self.g:
                for s in self.g[t].sources:
                    scm[t].append((s[0], -abs(s[1]))) 
        else:
            scm = {self.features.index(v): list() for v in self.features}
            for t in self.g:
                for s in self.g[t].sources:
                    scm[self.features.index(t)].append((self.features.index(s[0]), -abs(s[1]))) 
        return scm
    
    
    def get_Graph(self) -> dict:
        """
        Return Graph dictionary. E.g. {X1: {(X2, -2): '-->'}, X2: {(X3, -1): '-?>'}, X3: {(X3, -1): '-->'}}.

        Returns:
            dict: graph dictionary.
        """
        scm = {v: dict() for v in self.features}
        for t in self.g:
            for s in self.g[t].sources:
                scm[t][(s[0], -abs(s[1]))] = self.g[t].sources[s][TYPE] 
        return scm
    
        


"""
This module provides the DYNOTEARS class.

Classes:
    DYNOTEARS: class containing the DYNOTEARS causal discovery algorithm.
"""

from causalnex.structure.dynotears import from_pandas_dynamic
from causalflow.graph.DAG import DAG
from causalflow.causal_discovery.CausalDiscoveryMethod import CausalDiscoveryMethod 


class DYNOTEARS(CausalDiscoveryMethod):
    """DYNOTEARS causal discovery method."""
    
    def __init__(self, 
                 data, 
                 min_lag,
                 max_lag, 
                 verbosity, 
                 alpha = 0.05, 
                 resfolder = None,
                 neglect_only_autodep = False,
                 clean_cls = True):
        """
        Class constructor.

        Args:
            data (Data): data to analyse.
            min_lag (int): minimum time lag.
            max_lag (int): maximum time lag.
            verbosity (CPLevel): verbosity level.
            alpha (float, optional): PCMCI significance level. Defaults to 0.05.
            resfolder (string, optional): result folder to create. Defaults to None.
            neglect_only_autodep (bool, optional): Bit for neglecting variables with only autodependency. Defaults to False.
            clean_cls (bool): Clean console bit. Default to True.
        """
        super().__init__(data, min_lag, max_lag, verbosity, alpha, resfolder, neglect_only_autodep, clean_cls)
        
    def run(self) -> DAG:
        """
        Run DYNOTEARS algorithm.

        Returns:
            DAG: causal discovery result.
        """
        graph_dict = dict()
        for name in self.data.features:
            graph_dict[name] = []
        sm = from_pandas_dynamic(self.data.d, p=self.max_lag)

        tname_to_name_dict = dict()
        count_lag = 0
        idx_name = 0
        for tname in sm.nodes:
            tname_to_name_dict[tname] = self.data.features[idx_name]
            if count_lag == self.max_lag:
                idx_name = idx_name +1
                count_lag = -1
            count_lag = count_lag +1

        for ce in sm.edges:
            c = ce[0]
            e = ce[1]
            w = sm.adj[c][e]["weight"]
            tc = int(c.partition("lag")[2])
            te = int(e.partition("lag")[2])
            t = tc - te
            if (tname_to_name_dict[c], -t) not in graph_dict[tname_to_name_dict[e]]:
                graph_dict[tname_to_name_dict[e]].append((tname_to_name_dict[c], w, -t))

        self.CM = self._to_DAG(graph_dict)
        
        if self.resfolder is not None: self.logger.close()
        return self.CM
    
    
    def _to_DAG(self, graph):
        """
        Re-elaborate the result in a DAG.            

        Args:
            graph (dict): graph to convert into a DAG

        Returns:
            (DAG): result re-elaborated.
        """
        tmp_dag = DAG(self.data.features, self.min_lag, self.max_lag, self.neglect_only_autodep)
        tmp_dag.sys_context = dict()
        for t in graph.keys():
            for s in graph[t]:
                lag = abs(s[2])
                if lag >= self.min_lag and lag <= self.max_lag:
                    tmp_dag.add_source(t, s[0], abs(s[1]), 0, s[2])
        return tmp_dag


#!/usr/bin/env python
"""
Classes for artists used to display
- nodes / vertices
- edges
"""

import numpy as np
import matplotlib.pyplot as plt

from matplotlib.path import Path
from matplotlib.patches import PathPatch, transforms
from scipy.optimize import minimize_scalar
from matplotlib.patches import Circle

from ._utils import (
    _get_parallel_line,
    _get_orthogonal_unit_vector,
    _shorten_spline_by,
    _find_renderer,
    _get_radius,
)


class PathPatchDataUnits(PathPatch):
    """PathPatch in which the linewidth is also given in data units.

    Parameters
    ----------
    *args, **kwargs
        All arguments are passed through to PathPatch.

    Notes
    -----
    Adapted from https://stackoverflow.com/a/42972469/2912349

    """

    def __init__(self, *args, **kwargs):
        _lw_data = kwargs.pop("linewidth", 1)
        super().__init__(*args, **kwargs)
        self._lw_data = _lw_data

    def _get_lw(self):
        if self.axes is not None:
            ppd = 72./self.axes.figure.dpi
            trans = self.axes.transData.transform
            return ((trans((self._lw_data, self._lw_data))-trans((0, 0)))*ppd)[0]
            # return ((trans((self._lw_data, self._lw_data))-trans((0, 0)))*ppd)[1]
        else:
            return 1

    def _set_lw(self, lw):
        self._lw_data = lw

    _linewidth = property(_get_lw, _set_lw)


class NodeArtist(PathPatchDataUnits):
    """NodeArtist base class.

    Parameters
    ----------
    path :  matplotlib.Path instance
        The shape of the node.
        The path has to be closed and the vertices have to be centered on (0,0).
    xy : tuple
        The (float x, float y) coordinates of the centroid.
    size : float
        The maximum distance from the center to any of the vertices.
    **kwargs
        `Patch` properties:
        %(Patch_kwdoc)s

    See also
    --------
    :py:class:`CircularNodeArtist`, :py:class:`RegularPolygonArtist`

    """

    def __init__(self, path, xy, size, orientation=0, linewidth_correction=2, **kwargs):
        self._path = path
        self.xy = np.array(xy)
        self.size = size
        self.orientation = orientation
        self.linewidth_correction = linewidth_correction
        self.scale_by = 1 / _get_radius(self._path)
        self._patch_transform = transforms.Affine2D()
        super().__init__(path=self._path, **kwargs)
        self.transformed_path = self._path.transformed(self.get_patch_transform())

    def get_patch_transform(self):
        return self._patch_transform.clear() \
            .scale(self.scale_by * self.size - self._lw_data / self.linewidth_correction) \
            .rotate(self.orientation) \
            .translate(*self.xy)

    def _contains_points(self, points):
        return self.transformed_path.contains_points(points)

    def get_head_offset(self, edge_path):
        # Determine edge path points that are within node shape path.
        is_overlapping = self._contains_points(edge_path)

        if np.all(is_overlapping):
            import warnings
            warnings.warn("Node artist completely overlaps edge path!")
            return np.inf
        elif is_overlapping[-1]: # np.any(is_overlapping):
            # Resample segment consisting of last point that is not enclosed and first point that is.
            # idx = np.where(is_overlapping)[0][0] # This approach fails for self-loops.
            # segment = edge_path[[idx-1, idx]]
            idx = len(edge_path) - 1
            while is_overlapping[idx]:
                idx -= 1
            segment = edge_path[[idx, idx+1]]
            x, y = segment.T
            resampled = np.c_[np.linspace(x[0], x[1], 100), np.linspace(y[0], y[1], 100)]

            # Determine last resampled point that is not enclosed and compute distance to center.
            is_overlapping = self._contains_points(resampled)
            idx = np.where(is_overlapping)[0][0] - 1
            offset = np.linalg.norm(edge_path[-1] - resampled[idx])
            return offset
        else:
            import warnings
            warnings.warn("Node artist and edge path non-overlapping before offset correction!")
            return 0.

    def get_tail_offset(self, edge_path):
        return self.get_head_offset(edge_path[::-1])

    def get_maximum_fontsize(self, text_object, minimum=1, maximum=100):
        # NB: code assumes that
        # - fig.canvas.draw() has been called at least once
        # - fontdict contains parameters verticalalignment/va and horizontalalignment/ha and both are set to 'center'

        renderer = _find_renderer(text_object.get_figure())
        transform = self.axes.transData.inverted()

        def func(size_inverse):
            text_object.set_size(1/size_inverse)
            bbox_in_display_coordinates = text_object.get_window_extent(renderer)
            bbox_in_data_coordinates = bbox_in_display_coordinates.transformed(transform)
            if self.transformed_path.intersects_bbox(bbox_in_data_coordinates, filled=True) \
               and not self.transformed_path.intersects_bbox(bbox_in_data_coordinates, filled=False): # i.e. label fully enclosed
                return size_inverse
            else:
                return 1/minimum

        current_size = text_object.get_size()
        result = minimize_scalar(func, bounds=(1/maximum, 1/minimum))
        text_object.set_size(current_size)
        return 1 / result.x


class RegularPolygonNodeArtist(NodeArtist):
    """Instantiates a regular polygon node artist.

    Parameters
    ----------
    total_vertices : int
        Number of corners.
    orientation : float
        Orientation of the polygon in radians.
    xy : tuple
        The (float x, float y) coordinates of the centroid.
    size : float
        The distance from the center to each of the vertices.
    **kwargs
        `Patch` properties:
        %(Patch_kwdoc)s

    See also
    --------
    :py:class:`NodeArtist`, py:class:`CircularNodeArtist`

    """

    def __init__(self, total_vertices, orientation, xy, size, **kwargs):
        path = Path.unit_regular_polygon(total_vertices)
        linewidth_correction = 2 * np.sin(np.pi/total_vertices) # derives from the ratio between a side and the radius in a regular polygon.
        super().__init__(path, xy, size, orientation=orientation, linewidth_correction=linewidth_correction, **kwargs)


class CircularNodeArtist(NodeArtist):
    """Instantiates a circular node artist.

    Parameters
    ----------
    xy : tuple
        The (float x, float y) coordinates of the centroid.
    size : float
        The radius.
    **kwargs
        `Patch` properties:
        %(Patch_kwdoc)s

    See also
    --------
    py:class:`NodeArtist`, :py:class:`RegularPolygonArtist`

    """

    def __init__(self, xy, size, **kwargs):
        path = Path.circle()
        super().__init__(path, xy, size, **kwargs)

    def get_head_offset(self, edge_path):
        return self.size

    def get_tail_offset(self, edge_path):
        return self.size

    def get_maximum_fontsize(self, text_object, minimum=1, maximum=100):
        renderer = _find_renderer(text_object.get_figure())
        bbox_in_display_coordinates = text_object.get_window_extent(renderer)
        bbox_in_data_coordinates = bbox_in_display_coordinates.transformed(self.axes.transData.inverted())
        diameter = 2 * (self.size - self._lw_data/self.linewidth_correction)
        rescale_factor = diameter / np.sqrt(bbox_in_data_coordinates.width**2 + bbox_in_data_coordinates.height**2)
        current_size = text_object.get_size()
        return current_size * rescale_factor


class EdgeArtist(PathPatchDataUnits):
    """Implements the edge artist class.

    Parameters
    ----------
    midline : ndarray
        Array of (float x, float y) coordinates denoting the edge route.
    width : float, default 0.05
        The width of the edge (if shape is 'full').
    head_width : float, default 0.10
        Width of the arrow head.
        Set to a value close to zero (but not zero) to suppress drawing of arrowheads.
    head_length : float, default 0.15
        Length of the arrow head.
        Set to a value close to zero (but not zero) to suppress drawing of arrowheads.
    head_offset : float, default 0.
        For non-zero offset values, the end of the edge is offset from the target node.
        The distance is calculated along the midline.
    tail_offset : float, default 0.
        For non-zero offset values, the start of the edge is offset from the source node.
        The distance is calculated along the midline.
    shape : {'full', 'left', 'right'}, default 'full'
        The shape of the arrow.
        For shapes 'left' and 'right' the arrow only one half of the arrow is plotted.
    curved : bool, default False
        Indicates if the midline is straight (False) or curved (True).

    """
    def __init__(self, midline,
                 width       = 0.05,
                 head_width  = 0.10,
                 head_length = 0.15,
                 head_offset = 0.,
                 tail_offset = 0.,
                 shape       = 'full',
                 curved      = False,
                 *args, **kwargs):

        # book keeping
        self.midline     = midline
        self.width       = width
        self.head_width  = head_width
        self.head_length = head_length
        self.head_offset = head_offset
        self.tail_offset = tail_offset
        self.shape       = shape
        self.curved      = curved

        self._update_path() # sets self._path
        super().__init__(self._path, *args, **kwargs)


    def _update_path(self):
        # Determine the actual start (and hence midline) of the arrow given the tail offsets;
        # assume an ordered midline from source to target, i.e. from arrow base to arrow head.
        arrow_midline      = _shorten_spline_by(self.midline[::-1], self.tail_offset)[::-1]

        # Determine the actual endpoint (and hence midline) of the arrow given the offsets.
        arrow_midline      = _shorten_spline_by(arrow_midline, self.head_offset)
        arrow_tail_midline = _shorten_spline_by(arrow_midline, self.head_length)

        head_vertex_tip  = arrow_midline[-1]
        head_vertex_base = arrow_tail_midline[-1]
        (dx, dy), = _get_orthogonal_unit_vector(np.atleast_2d(head_vertex_tip - head_vertex_base)) * self.head_width / 2.

        if self.shape == 'full':
            tail_vertices_right = _get_parallel_line(arrow_tail_midline, -self.width / 2.)
            tail_vertices_left  = _get_parallel_line(arrow_tail_midline,  self.width / 2.)
            head_vertex_right = head_vertex_base - np.array([dx, dy])
            head_vertex_left  = head_vertex_base + np.array([dx, dy])

            vertices = np.concatenate([
                tail_vertices_right[::-1],
                tail_vertices_left,
                head_vertex_left[np.newaxis,:],
                head_vertex_tip[np.newaxis,:],
                head_vertex_right[np.newaxis,:],
                tail_vertices_right[np.newaxis,-1],
            ])
            codes = np.concatenate([
                [Path.MOVETO] + [Path.LINETO for _ in tail_vertices_right[1:]],
                [Path.LINETO for _ in tail_vertices_left],
                [Path.LINETO], # head_vertex_left
                [Path.LINETO], # head_vertex_tip
                [Path.LINETO], # head_vertex_right
                [Path.CLOSEPOLY] # tail_vertices_right[-1]
            ])

        elif self.shape == 'right':
            # tail_vertices_right = _get_parallel_line(arrow_tail_midline, -self.width / 2.)
            tail_vertices_right = _get_parallel_line(arrow_tail_midline, -0.6 * self.width)
            arrow_tail_midline = _get_parallel_line(arrow_tail_midline, -0.1 * self.width)
            head_vertex_right  = head_vertex_base - np.array([dx, dy])

            vertices = np.concatenate([
                tail_vertices_right[::-1],
                arrow_tail_midline,
                head_vertex_tip[np.newaxis,:],
                head_vertex_right[np.newaxis,:],
                tail_vertices_right[np.newaxis,-1]
            ])
            codes = np.concatenate([
                [Path.MOVETO] + [Path.LINETO for _ in tail_vertices_right[1:]],
                [Path.LINETO for _ in arrow_tail_midline],
                [Path.LINETO], # head_vertex_tip
                [Path.LINETO], # head_vertex_right
                [Path.CLOSEPOLY] # tail_vertices_right[-1]
            ])

        elif self.shape == 'left':
            # tail_vertices_left = _get_parallel_line(arrow_tail_midline,  self.width / 2.)
            tail_vertices_left = _get_parallel_line(arrow_tail_midline,  0.6 * self.width)
            arrow_tail_midline = _get_parallel_line(arrow_tail_midline,  0.1 * self.width)
            head_vertex_left = head_vertex_base + np.array([dx, dy])

            vertices = np.concatenate([
                arrow_tail_midline[::-1],
                tail_vertices_left,
                head_vertex_left[np.newaxis,:],
                head_vertex_tip[np.newaxis,:],
                arrow_tail_midline[np.newaxis,-1],
            ])
            codes = np.concatenate([
                [Path.MOVETO] + [Path.LINETO for _ in arrow_tail_midline[1:]],
                [Path.LINETO for _ in tail_vertices_left],
                [Path.LINETO], # head_vertex_left
                [Path.LINETO], # head_vertex_tip
                [Path.CLOSEPOLY] # arrow_tail_midline[-1]
            ])

        else:
            raise ValueError("Argument 'shape' needs to one of: 'left', 'right', 'full', not '{}'.".format(self.shape))

        self._path = Path(vertices, codes)


    def update_midline(self, midline):
        """Update the midline and recompute the edge path."""
        self.midline = midline
        self._update_path()


    def update_width(self, width, arrow=True):
        """
        Adjust the edge width. If arrow is True, the arrow head length and
        width are rescaled by the ratio new width / old width.
        """
        if arrow:
            ratio = width / self.width
            self.head_length *= ratio
            self.head_width *= ratio
        self.width = width
        self._update_path()


    # def draw_tail_marker(self, ax, tail_marker, tails_size, tail_marker_color, tail_marker_bordercolor, node_radius):
    #     tail_vertex_base = self.midline[0]
    #     direction_vector = self.midline[1] - self.midline[0]
    #     direction_vector /= np.linalg.norm(direction_vector)  # Normalize direction vector

    #     # Offset tail marker position by node radius in the opposite direction of the arrow
    #     tail_marker_position = tail_vertex_base + direction_vector * node_radius
    #     # ax.scatter(tail_marker_position[0], tail_marker_position[1], marker=tail_marker, color=tail_marker_color, edgecolors=tail_marker_bordercolor, s=tails_size, zorder=1000)
    #     circle = Circle(tail_marker_position, tails_size, facecolor=tail_marker_color, edgecolor=tail_marker_bordercolor, zorder=1000)
    #     ax.add_patch(circle)
        
    # def draw_head_marker(self, ax, head_marker, head_size, head_marker_color, head_marker_bordercolor, node_radius):
    #     head_vertex_base = self.midline[-1]
    #     direction_vector = self.midline[-1] - self.midline[-2]
    #     direction_vector /= np.linalg.norm(direction_vector)  # Normalize direction vector

    #     # Offset tail marker position by node radius in the opposite direction of the arrow
    #     head_marker_position = head_vertex_base - direction_vector * node_radius
    #     # ax.scatter(head_marker_position[0], head_marker_position[1], marker=head_marker, color=head_marker_color, edgecolors=head_marker_bordercolor, s=head_size, zorder=1000)
    #     circle = Circle(head_marker_position, head_size, facecolor=head_marker_color, edgecolor=head_marker_bordercolor, zorder=1000)
    #     ax.add_patch(circle)
        
    def draw_marker(self, ax, marker_type, marker, marker_size, marker_color, marker_edgecolor, node_radius):
        if marker_type == 'h':
            vertex_base = self.midline[-1]
            direction_vector = self.midline[-1] - self.midline[-2]
            direction_vector /= np.linalg.norm(direction_vector)
            marker_position = vertex_base - direction_vector * node_radius
        else:
            vertex_base = self.midline[0]
            direction_vector = self.midline[1] - self.midline[0]
            direction_vector /= np.linalg.norm(direction_vector)
            marker_position = vertex_base + direction_vector * node_radius
            
        
        if marker == 'o':
            # ax.scatter(head_marker_position[0], head_marker_position[1], marker=head_marker, color=head_marker_color, edgecolors=head_marker_bordercolor, s=head_size, zorder=1000)
            circle = Circle(marker_position, marker_size, facecolor=marker_color, edgecolor=marker_edgecolor, zorder=1000)
            ax.add_patch(circle)
        elif marker == 'x':
            ax.scatter(marker_position[0], marker_position[1], marker='x', color=marker_color, edgecolors=marker_edgecolor, s=10, zorder=1000)
        else:
            raise ValueError("Unsupported marker type. Use 'o' or 'x'")

            


"""
This module provides various classes for Mutual Information-based feature selection analysis.

Classes:
    MIestimator: support class for handling different Mutual Information estimators.
    MI: Mutual Information class.
"""

from enum import Enum
from causalflow.selection_methods.SelectionMethod import SelectionMethod, CTest, _suppress_stdout
from idtxl.multivariate_mi import MultivariateMI
from idtxl.data import Data
from fpcmci.CPrinter import CP
from scipy.stats import shapiro, kstest
import importlib

class MIestimator(Enum):
    """MIestimator Enumerator."""

    Auto = 'Auto'
    Gaussian = 'JidtGaussianCMI'
    Kraskov = 'JidtKraskovCMI'
    OpenCLKraskov = 'OpenCLKraskovCMI'


class MI(SelectionMethod):
    """Feature selection method based on Mutual Information analysis."""
    
    def __init__(self, estimator: MIestimator):
        """
        Class contructor.

        Args:
            estimator (MIestimator): Gaussian/Kraskov
        """
        super().__init__(CTest.MI)
        self.estimator = estimator
        
    @property
    def isOpenCLinstalled(self) -> bool:
        """
        Check whether the pyopencl pkg is installed.

        Returns:
            bool: True if pyopencl is installed.
        """
        try:
            importlib.import_module('pyopencl')
            return True
        except ImportError:
            return False
        
    def _select_estimator(self):
        """Select the MI estimator."""
        CP.info("\n##")
        CP.info("## MI Estimator selection")
        CP.info("##")

        isGaussian = True

        for f in self.data.features:
            # Perform Shapiro-Wilk test
            shapiro_stat, shapiro_p_value = shapiro(self.data.d[f])
            # Perform Kolmogorov-Smirnov test
            ks_stat, ks_p_value = kstest(self.data.d[f], 'norm')
                
            # Print results
            CP.debug("\n")
            CP.debug(f"Feature '{f}':")
            CP.debug(f"\t- Shapiro-Wilk test: val={round(shapiro_stat, 2)}, p-val={round(shapiro_p_value, 2)}")
            CP.debug(f"\t- Kolmogorov-Smirnov test: val={round(ks_stat, 2)}, p-val={round(ks_p_value, 2)}")
                
            # Check if p-values are less than significance level (e.g., 0.05) for normality
            if shapiro_p_value < 0.05 or ks_p_value < 0.05:
                CP.debug("\tNot normally distributed")
                isGaussian = False
                # break
            else:
                CP.debug("\tNormally distributed")
                
        if isGaussian:
            self.estimator = MIestimator.Gaussian
        else:
            self.estimator = MIestimator.OpenCLKraskov if self.isOpenCLinstalled else MIestimator.Kraskov
        CP.info("\n## MI Estimator: " + self.estimator.value)

    def compute_dependencies(self):
        """
        Compute list of dependencies for each target by mutual information analysis.

        Returns:
            (DAG): dependency dag
        """
        if self.estimator is MIestimator.Auto: self._select_estimator()

        with _suppress_stdout():
            data = Data(self.d.values, dim_order='sp') # sp = samples(row) x processes(col)

            network_analysis = MultivariateMI()
            settings = {'cmi_estimator': self.estimator.value,
                        'max_lag_sources': self.max_lag,
                        'min_lag_sources': self.min_lag,
                        'alpha_max_stats': self.alpha,
                        'alpha_min_stats': self.alpha,
                        'alpha_omnibus': self.alpha,
                        'alpha_max_seq': self.alpha,
                        'verbose': False}
            results = network_analysis.analyse_network(settings=settings, data=data)

        for t in results._single_target.keys():
            sel_sources = [s[0] for s in results._single_target[t]['selected_vars_sources']]
            if sel_sources:
                sel_sources_lag = [s[1] for s in results._single_target[t]['selected_vars_sources']]
                sel_sources_score = results._single_target[t]['selected_sources_mi']
                sel_sources_pval = results._single_target[t]['selected_sources_pval']
                for s, score, pval, lag in zip(sel_sources, sel_sources_score, sel_sources_pval, sel_sources_lag):
                    self._add_dependency(self.features[t], self.features[s], score, pval, lag)

        return self.result